import networkx as nx
from typing import Sequence, MutableSet, Optional
from pyomo.contrib.fbbt.fbbt import fbbt, compute_bounds_on_expr
import time
import enum
import warnings
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from .obbt import perform_obbt as normal_obbt
from .filters import aggressive_filter
import pyomo.environ as pe
from pyomo.core.expr.visitor import identify_variables
from pyomo.common.collections import ComponentSet
from pyomo.common.collections.orderedset import OrderedSet
from coramin.relaxations.iterators import relaxation_data_objects, nonrelaxation_component_data_objects
from pyomo.core.expr.visitor import replace_expressions
import logging
import networkx
import metis
import numpy as np
import math
from pyomo.core.base.block import declare_custom_block, _BlockData
from coramin.utils.pyomo_utils import get_objective
from pyomo.core.base.var import _GeneralVarData
from coramin.relaxations.copy_relaxation import copy_relaxation_with_local_data
from coramin.relaxations.relaxations_base import BaseRelaxationData
from coramin.utils import RelaxationSide
from collections import defaultdict
from pyomo.core.expr import numeric_expr
from pyomo.core.expr.visitor import ExpressionValueVisitor, nonpyomo_leaf_types
from pyomo.common.modeling import unique_component_name
from coramin.relaxations.split_expr import flatten_expr


logger = logging.getLogger(__name__)


class DecompositionError(Exception):
    pass


class TreeBlockError(DecompositionError):
    pass


@declare_custom_block(name='TreeBlock')
class TreeBlockData(_BlockData):
    def __init__(self, component):
        _BlockData.__init__(self, component)
        self._children_index = None
        self._children = None
        self._linking_constraints = None
        self._already_setup = False
        self._is_leaf = None
        self._is_root = True
        self._allow_changes = False

    def setup(self, children_keys):
        assert not self._already_setup
        self._already_setup = True
        if len(children_keys) == 0:
            self._is_leaf = True
        else:
            self._is_leaf = False
            del self._children_index
            del self._children
            del self._linking_constraints
            self._allow_changes = True
            self._children_index = pe.Set(initialize=children_keys)
            self._children = TreeBlock(self._children_index)
            self._linking_constraints = pe.ConstraintList()
            self._allow_changes = False
            for key in children_keys:
                child = self.children[key]
                child._is_root = False

    def _assert_setup(self):
        if not self._already_setup:
            raise TreeBlockError('The TreeBlock has not been setup yet. Please call the setup method.')

    def is_leaf(self):
        self._assert_setup()
        return self._is_leaf

    def add_component(self, name, val):
        self._assert_setup()
        if self.is_leaf() or self._allow_changes or not isinstance(val, _GeneralVarData):
            _BlockData.add_component(self, name, val)
        else:
            raise TreeBlockError('Pyomo variables cannot be added to a TreeBlock unless it is a leaf.')

    @property
    def children(self):
        self._assert_setup()
        if self.is_leaf():
            raise TreeBlockError('Leaf TreeBlocks do not have children. Please check the is_leaf method')
        return self._children

    @property
    def linking_constraints(self):
        self._assert_setup()
        if self.is_leaf():
            raise TreeBlockError('leaf TreeBlocks do not have linking_constraints. Please check the is_leaf method.')
        return self._linking_constraints

    def _num_stages(self):
        self._assert_setup()
        num_stages = 1
        if not self.is_leaf():
            num_stages += max([child._num_stages() for child in self.children.values()])
        return num_stages

    def num_stages(self):
        if not self._is_root:
            raise TreeBlockError('The num_stages method can only be called from the root TreeBlock')
        return self._num_stages()

    @staticmethod
    def _stage_blocks(children, count, stage):
        if count == stage:
            for child in children.values():
                yield child
        else:
            for child in children.values():
                if not child.is_leaf():
                    for b in TreeBlockData._stage_blocks(child.children, count+1, stage):
                        yield b

    def stage_blocks(self, stage, active=None):
        self._assert_setup()
        if not self._is_root:
            raise TreeBlockError('The num_stages method can only be called from the root TreeBlock')
        if stage == 0:
            if (active and self.active) or (not active):
                yield self
        elif not self.is_leaf():
            for b in self._stage_blocks(self.children, 1, stage):
                if (active and b.active) or (not active):
                    yield b

    def get_block_stage(self, block):
        self._assert_setup()
        if not self._is_root:
            raise TreeBlockError('The get_block_stage method can only be called from the root TreeBlock.')
        for stage_ndx in range(self.num_stages()):
            stage_blocks = OrderedSet(self.stage_blocks(stage_ndx))
            if block in stage_blocks:
                return stage_ndx
        return None


class _Node(object):
    def __init__(self, comp):
        self.comp = comp

    def is_var(self):
        return False

    def is_con(self):
        return False

    def is_rel(self):
        return False

    def __repr__(self):
        return str(self.comp)

    def __str__(self):
        return str(self.comp)

    def __eq__(self, other):
        if isinstance(other, _Node):
            return self.comp is other.comp
        return False

    def __hash__(self):
        return hash(id(self.comp))


class _VarNode(_Node):
    def is_var(self):
        return True


class _ConNode(_Node):
    def is_con(self):
        return True


class _RelNode(_Node):
    def is_rel(self):
        return True


class _Edge(object):
    def __init__(self, node1: _VarNode, node2: _Node):
        assert node1.is_var()
        self.node1 = node1
        self.node2 = node2

    def __str__(self):
        s = 'Edge from {0} to {1}'.format(str(self.node1), str(self.node2))
        return s


class _Tree(object):
    def __init__(self, children=None, edges_between_children=None):
        """
        Parameters
        ----------
        children: list or collections.abc.Iterable of _Tree or networkx.Graph
        edges_between_children: list or collections.abc.Iterable of _Edge
        """
        self.children = OrderedSet()
        self.edges_between_children = OrderedSet()
        if children is not None:
            self.children.update(children)
        if edges_between_children is not None:
            self.edges_between_children.update(edges_between_children)

    def build_pyomo_model(self, block):
        """
        Parameters
        ----------
        block: TreeBlockData
            empty TreeBlock

        Returns
        -------
        component_map: pe.ComponentMap
        """
        block.setup(children_keys=list(range(len(self.children))))
        component_map = pe.ComponentMap()
        replacement_map_by_child = dict()

        for i, child in enumerate(self.children):
            if isinstance(child, _Tree):
                tmp_component_map = child.build_pyomo_model(block=block.children[i])
            elif isinstance(child, networkx.Graph):
                block.children[i].setup(children_keys=list())
                tmp_component_map = build_pyomo_model_from_graph(graph=child, block=block.children[i])
            else:
                raise ValueError('Unexpected child type: {0}'.format(str(type(child))))
            replacement_map_by_child[child] = tmp_component_map
            component_map.update(tmp_component_map)

        logger.debug('creating linking cons linking the children of {0}'.format(str(block)))
        for edge in self.edges_between_children:
            logger.debug('adding linking constraint for edge {0}'.format(str(edge)))
            if edge.node1.comp is not edge.node2.comp:
                raise DecompositionError('Edge {0} node1.comp is not node2.comp'.format(edge))
            if edge.node1.comp not in component_map:
                logger.warning('Edge {0} node {1} is not in the component map'.format(str(edge), str(edge.node1)))
            all_children = list(self.children)
            assert len(all_children) == 2
            child0 = all_children[0]
            child1 = all_children[1]
            v1 = replacement_map_by_child[child0][edge.node1.comp]
            v2 = replacement_map_by_child[child1][edge.node2.comp]
            assert v1 is not v2
            block.linking_constraints.add(v1 == v2)

        return component_map

    def log(self, prefix=''):
        logger.debug(prefix + '# Edges: {0}'.format(len(self.edges_between_children)))
        for _child in self.children:
            if isinstance(_child, _Tree):
                _child.log(prefix=prefix + '  ')
            else:
                logger.debug(prefix + '  Leaf: # NNZ: {0}'.format(_child.number_of_edges()))


def _is_dominated(ndx, num_cuts, balance, num_cuts_array, balance_array):
    cut_diff = ((num_cuts - num_cuts_array) >= 0)
    balance_diff = ((abs(balance - 0.5) - abs(balance_array - 0.5)) >= 0)
    cut_diff[ndx] = False
    balance_diff[ndx] = False
    return np.any(cut_diff & balance_diff)


def _networkx_to_adjacency_list(graph: networkx.Graph):
    adj_list = list()
    node_to_ndx_map = dict()
    for ndx, node in enumerate(graph.nodes):
        node_to_ndx_map[node] = ndx

    for ndx, node in enumerate(graph.nodes):
        adj_list.append(list())
        for other_node in graph.adj[node].keys():
            other_ndx = node_to_ndx_map[other_node]
            adj_list[ndx].append(other_ndx)

    return adj_list


def choose_metis_partition(graph, max_size_diff_trials, seed_trials):
    """
    Parameters
    ----------
    graph: networkx.Graph
    max_size_diff_trials: list of float
    seed_trials: list of int

    Returns
    -------
    max_size_diff_selected: float
    seed_selected: float
    """
    cut_list = list()
    for _max_size_diff in max_size_diff_trials:
        for _seed in seed_trials:
            if _seed is None:
                edgecuts, parts = metis.part_graph(_networkx_to_adjacency_list(graph), nparts=2, ubvec=[1 + _max_size_diff])
            else:
                edgecuts, parts = metis.part_graph(_networkx_to_adjacency_list(graph), nparts=2, ubvec=[1 + _max_size_diff], seed=_seed)
            cut_list.append((edgecuts, sum(parts)/graph.number_of_nodes(), _max_size_diff, _seed))
    cut_list.sort(key=lambda i: i[0])

    ############################
    # get the "pareto front" obtained with metis
    ############################
    num_cuts_array = np.array([i[0] for i in cut_list])
    balance_array = np.array([i[1] for i in cut_list])

    pareto_list = list()
    for ndx, partition in enumerate(cut_list):
        num_cuts = partition[0]
        balance = partition[1]
        if not _is_dominated(ndx, num_cuts, balance, num_cuts_array, balance_array):
            pareto_list.append(partition)
    if len(pareto_list) == 0:
        pareto_list.append(cut_list[0])

    selection = 0
    chosen_partition = pareto_list[selection]
    max_size_diff_selected = chosen_partition[2]
    seed_selected = chosen_partition[3]
    return max_size_diff_selected, seed_selected


def evaluate_partition(original_graph, tree):
    """
    Parameters
    ----------
    original_graph: networkx.Graph
    tree: _Tree
    """
    original_graph_nnz = original_graph.number_of_edges()
    original_graph_n_vars_to_tighten = len(collect_vars_to_tighten_from_graph(graph=original_graph))
    original_obbt_nnz = original_graph_nnz * original_graph_n_vars_to_tighten

    tree_obbt_nnz = 0
    tree_nnz = 0
    assert len(tree.children) == 2
    for child in tree.children:
        assert isinstance(child, networkx.Graph)
        child_nnz = child.number_of_edges()
        tree_nnz += child_nnz
        child_n_vars_to_tighten = len(collect_vars_to_tighten_from_graph(graph=child))
        tree_obbt_nnz += child_nnz * child_n_vars_to_tighten
    tree_nnz += 2 * len(tree.edges_between_children)
    tree_obbt_nnz += tree_nnz * len(tree.edges_between_children)
    partitioning_ratio = original_obbt_nnz / tree_obbt_nnz
    return partitioning_ratio


def _refine_partition(graph: nx.Graph, model: _BlockData,
                      removed_edges: Sequence[_Edge],
                      graph_a_nodes: MutableSet[_Node],
                      graph_b_nodes: MutableSet[_Node]):
    con_count = defaultdict(int)
    for edge in removed_edges:
        n1, n2 = edge.node1, edge.node2
        if n1.is_con():
            con_count[n1.comp] += 1
        if n2.is_con():
            con_count[n2.comp] += 1

    for c, count in con_count.items():
        if count < 3:
            continue

        new_body = flatten_expr(c.body)

        if type(new_body) is not numeric_expr.SumExpression:
            logger.info(f'Constraint {str(c)} is contributing to {count} removed '
                        f'edges, but we cannot split the constraint because the '
                        f'body is not a SumExpression.')
            continue

        graph_a_args = list()
        graph_b_args = list()
        correct_structure = True
        for arg in new_body.args:
            graph_a_arg_vars = ComponentSet()
            graph_b_arg_vars = ComponentSet()
            for v in identify_variables(arg, include_fixed=False):
                v_node = _VarNode(v)
                assert v_node in graph_a_nodes or v_node in graph_b_nodes
                if v_node in graph_a_nodes:
                    graph_a_arg_vars.add(v)
                else:
                    graph_b_arg_vars.add(v)
            if len(graph_a_arg_vars) > 0 and len(graph_b_arg_vars) > 0:
                correct_structure = False
                break
            if len(graph_a_arg_vars) > 0:
                graph_a_args.append(arg)
            elif len(graph_b_arg_vars) > 0:
                graph_b_args.append(arg)
            else:
                graph_a_args.append(arg)

        if not correct_structure:
            logger.info(f'Constriant {str(c)} is contributing to {count} removed '
                        f'edges, but we cannot split the constraint because some of '
                        f'the terms in the SumExpression contain variables from both '
                        f'partitions.')
            continue

        # update the model
        if not hasattr(model, 'dbt_partition_vars'):
            model.dbt_partition_vars = pe.VarList()
            model.dbt_partition_cons = pe.ConstraintList()

        graph_a_var = model.dbt_partition_vars.add()
        graph_b_var = model.dbt_partition_vars.add()

        if c.lower is not None and c.upper is not None:
            new_c1 = model.dbt_partition_cons.add(graph_a_var == sum(graph_a_args))
            new_c2 = model.dbt_partition_cons.add(graph_b_var == sum(graph_b_args))
            if c.equality:
                new_c3 = model.dbt_partition_cons.add(graph_a_var + graph_b_var == c.lower)
            else:
                new_c3 = model.dbt_partition_cons.add((c.lower, graph_a_var + graph_b_var, c.upper))
        elif c.lower is None:
            assert c.upper is not None
            new_c1 = model.dbt_partition_cons.add(graph_a_var >= sum(graph_a_args))
            new_c2 = model.dbt_partition_cons.add(graph_b_var >= sum(graph_b_args))
            new_c3 = model.dbt_partition_cons.add(graph_a_var + graph_b_var <= c.upper)
        else:
            assert c.upper is None
            new_c1 = model.dbt_partition_cons.add(graph_a_var <= sum(graph_a_args))
            new_c2 = model.dbt_partition_cons.add(graph_b_var <= sum(graph_b_args))
            new_c3 = model.dbt_partition_cons.add(graph_a_var + graph_b_var >= c.lower)
        c.deactivate()

        # update the graph
        graph.remove_node(_ConNode(c))
        graph.add_node(_VarNode(graph_a_var))
        graph.add_node(_VarNode(graph_b_var))
        for new_con in [new_c1, new_c2, new_c3]:
            graph.add_node(_ConNode(new_con))
            for v in identify_variables(new_con.body, include_fixed=False):
                graph.add_edge(_VarNode(v), _ConNode(new_con))

        # update removed_edges
        new_removed_edges = list()
        for e in removed_edges:
            if e.node2.comp is not c:
                new_removed_edges.append(e)

        new_removed_edges.append(_Edge(_VarNode(graph_a_var), _ConNode(new_c3)))
        removed_edges = new_removed_edges

        # update graph_a_nodes and graph_b_nodes
        graph_a_nodes.discard(_ConNode(c))
        graph_b_nodes.discard(_ConNode(c))
        graph_a_nodes.add(_VarNode(graph_a_var))
        graph_b_nodes.add(_VarNode(graph_b_var))
        graph_a_nodes.add(_ConNode(new_c1))
        graph_b_nodes.add(_ConNode(new_c2))
        graph_b_nodes.add(_ConNode(new_c3))

    return removed_edges


def split_metis(graph, model):
    """
    Parameters
    ----------
    graph: networkx.Graph
    model: _BlockData

    Returns
    -------
    tree: _Tree
    """
    max_size_diff, seed = choose_metis_partition(graph, max_size_diff_trials=[0.15], seed_trials=list(range(10)))
    if seed is None:
        edgecuts, parts = metis.part_graph(_networkx_to_adjacency_list(graph), nparts=2, ubvec=[1 + max_size_diff])
    else:
        edgecuts, parts = metis.part_graph(_networkx_to_adjacency_list(graph), nparts=2, ubvec=[1 + max_size_diff], seed=seed)

    graph_a_nodes = OrderedSet()
    graph_b_nodes = OrderedSet()
    for ndx, n in enumerate(graph.nodes()):
        if parts[ndx] == 0:
            graph_a_nodes.add(n)
        else:
            assert parts[ndx] == 1
            graph_b_nodes.add(n)

    removed_edges = list()
    for n1, n2 in graph.edges():
        if not n1.is_var():
            assert n2.is_var()
            n1, n2 = n2, n1
        else:
            assert not n2.is_var()
        if n1 in graph_a_nodes and n2 in graph_a_nodes:
            continue
        elif n1 in graph_b_nodes and n2 in graph_b_nodes:
            continue
        else:
            removed_edges.append(_Edge(n1, n2))

    removed_edges = _refine_partition(graph=graph, model=model,
                                      removed_edges=removed_edges,
                                      graph_a_nodes=graph_a_nodes,
                                      graph_b_nodes=graph_b_nodes)

    graph_a_edges = list()
    graph_b_edges = list()
    for n1, n2 in graph.edges():
        if not n1.is_var():
            assert n2.is_var()
            n1, n2 = n2, n1
        else:
            assert not n2.is_var()
        if n1 in graph_a_nodes and n2 in graph_a_nodes:
            graph_a_edges.append((n1, n2))
        elif n1 in graph_b_nodes and n2 in graph_b_nodes:
            graph_b_edges.append((n1, n2))
        else:
            continue

    linking_edges = list()
    new_var_nodes_dict = dict()
    for e in removed_edges:
        n1, n2 = e.node1, e.node2
        assert n1.is_var()
        assert not n2.is_var()
        if n1 in new_var_nodes_dict:
            new_var_node = new_var_nodes_dict[n1]
        else:
            new_var_node = _VarNode(n1.comp)
            new_var_nodes_dict[n1] = new_var_node
            linking_edge = _Edge(n1, new_var_node)
            linking_edges.append(linking_edge)
        if n1 in graph_a_nodes:
            assert n2 in graph_b_nodes
            graph_b_edges.append((new_var_node, n2))
        else:
            assert n1 in graph_b_nodes
            assert n2 in graph_a_nodes
            graph_a_edges.append((new_var_node, n2))

    graph_a = networkx.Graph()
    graph_b = networkx.Graph()

    graph_a.add_nodes_from(graph_a_nodes)
    graph_b.add_nodes_from(graph_b_nodes)
    graph_a.add_edges_from(graph_a_edges)
    graph_b.add_edges_from(graph_b_edges)

    if ((graph_a.number_of_nodes() >= 0.99 * graph.number_of_nodes()) or
            (graph_b.number_of_nodes() >= 0.99 * graph.number_of_nodes())):
        raise DecompositionError('Failed to partition graph')

    tree = _Tree(children=[graph_a, graph_b], edges_between_children=linking_edges)

    partitioning_ratio = evaluate_partition(original_graph=graph, tree=tree)

    return tree, partitioning_ratio


def convert_pyomo_model_to_bipartite_graph(m: _BlockData):
    """
    Parameters
    ----------
    m: _BlockData

    Returns
    -------
    graph: networkx.Graph
    """
    graph = networkx.Graph()
    var_map = pe.ComponentMap()

    for v in nonrelaxation_component_data_objects(m, pe.Var, sort=True, descend_into=True):
        if v.fixed:
            continue
        var_map[v] = _VarNode(v)
        graph.add_node(var_map[v])

    for b in relaxation_data_objects(m, descend_into=True, active=True, sort=True):
        node2 = _RelNode(b)
        for v in (list(b.get_rhs_vars()) + [b.get_aux_var()]):
            node1 = var_map[v]
            graph.add_edge(node1, node2)

    for c in nonrelaxation_component_data_objects(m, pe.Constraint, active=True, sort=True, descend_into=True):
        node2 = _ConNode(c)
        for v in identify_variables(c.body, include_fixed=False):
            node1 = var_map[v]
            graph.add_edge(node1, node2)

    return graph


def build_pyomo_model_from_graph(graph, block):
    """
    Parameters
    ----------
    graph: networkx.Graph
    block: pe.Block

    Returns
    -------
    component_map: pe.ComponentMap
    """
    vars = list()
    cons = list()
    rels = list()
    var_names = list()
    con_names = list()
    rel_names = list()
    for node in graph.nodes():
        if node.is_var():
            vars.append(node)
            var_names.append(node.comp.getname(fully_qualified=True).replace('.', '_'))
        elif node.is_con():
            cons.append(node)
            con_names.append(node.comp.getname(fully_qualified=True).replace('.', '_'))
        else:
            assert node.is_rel()
            rels.append(node)
            rel_names.append(node.comp.getname(fully_qualified=True).replace('.', '_'))

    assert len(vars) == len(set(vars))
    assert len(cons) == len(set(cons))
    assert len(rels) == len(set(rels))

    block.var_names = pe.Set(initialize=var_names)
    block.con_names = pe.Set(initialize=con_names)
    block.vars = pe.Var(block.var_names)
    block.cons = pe.Constraint(block.con_names)
    block.rels = pe.Block()

    component_map = pe.ComponentMap()
    for v_name, v in zip(var_names, vars):
        new_v = block.vars[v_name]
        component_map[v.comp] = new_v
        new_v.setlb(v.comp.lb)
        new_v.setub(v.comp.ub)
        new_v.domain = v.comp.domain
        if v.comp.is_fixed():
            new_v.fix(v.comp.value)
        new_v.set_value(v.comp.value, skip_validation=True)

    var_map = {id(k): v for k, v in component_map.items()}

    for c_name, c in zip(con_names, cons):
        if c.comp.equality:
            block.cons[c_name] = (replace_expressions(c.comp.body, substitution_map=var_map,
                                                      remove_named_expressions=True) == c.comp.lower)
        else:
            block.cons[c_name] = (pe.inequality(lower=c.comp.lower,
                                                body=replace_expressions(c.comp.body, substitution_map=var_map,
                                                                         remove_named_expressions=True),
                                                upper=c.comp.upper))
        component_map[c.comp] = block.cons[c_name]

    for r_name, r in zip(rel_names, rels):
        new_rel = copy_relaxation_with_local_data(r.comp, var_map)
        setattr(block.rels, r_name, new_rel)
        new_rel.rebuild()
        component_map[r.comp] = new_rel

    return component_map


def num_cons_in_graph(graph, include_rels=True):
    res = 0

    if include_rels:
        for n in graph.nodes():
            if n.is_con() or n.is_rel():
                res += 1
    else:
        for n in graph.nodes():
            if n.is_con():
                res += 1

    return res


class DecompositionStatus(enum.Enum):
    normal = 0  # the model was successfullay decomposed at least once and no exception was raised
    error = 1  # an exception was raised
    bad_ratio = 2  # the model could not be decomposed at all because the min_parition_ratio was not satisfied
    problem_too_small = 3  # the model could not be decomposed at all because the number of jacobian nonzeros in the original problem was less than max_leaf_nnz


def compute_partition_ratio(original_model: _BlockData, decomposed_model: TreeBlockData):
    graph = convert_pyomo_model_to_bipartite_graph(original_model)
    pr_numerator = graph.number_of_edges() * len(collect_vars_to_tighten(original_model))

    pr_denominator = 0
    vars_to_tighten_by_block = collect_vars_to_tighten_by_block(decomposed_model, 'dbt')
    for block, vars_to_tighten in vars_to_tighten_by_block.items():
        pr_denominator += len(vars_to_tighten) * convert_pyomo_model_to_bipartite_graph(block).number_of_edges()
    pr = pr_numerator / pr_denominator
    return pr


def _reformulate_objective(model):
    current_obj = get_objective(model)
    if current_obj is None:
        raise ValueError('No active objective found!')
    if not current_obj.expr.is_variable_type():
        obj_var_name = unique_component_name(model, 'obj_var')
        obj_var = pe.Var(bounds=compute_bounds_on_expr(current_obj.expr))
        model.add_component(obj_var_name, obj_var)
        model.del_component(current_obj)
        new_objective = pe.Objective(expr=model.obj_var)
        new_obj_name = unique_component_name(model, 'objective')
        model.add_component(new_obj_name, new_objective)
        if current_obj.sense == pe.minimize:
            obj_con = pe.Constraint(expr=current_obj.expr <= obj_var)
        else:
            obj_con = pe.Constraint(expr=current_obj.expr >= obj_var)
            new_objective.sense = pe.maximize
        obj_con_name = unique_component_name(model, 'obj_con')
        model.add_component(obj_con_name, obj_con)


def _eliminate_mutable_params(model):
    sub_map = dict()
    for p in nonrelaxation_component_data_objects(model, pe.Param, descend_into=True):
        sub_map[id(p)] = p.value

    for c in nonrelaxation_component_data_objects(model, pe.Constraint, active=True, descend_into=True):
        if c.lower is None:
            new_lower = None
        else:
            new_lower = replace_expressions(c.lower, sub_map,
                                            descend_into_named_expressions=True,
                                            remove_named_expressions=True)
        new_body = replace_expressions(c.body, sub_map,
                                       descend_into_named_expressions=True,
                                       remove_named_expressions=True)
        if c.upper is None:
            new_upper = None
        else:
            new_upper = replace_expressions(c.upper, sub_map,
                                            descend_into_named_expressions=True,
                                            remove_named_expressions=True)
        c.set_value((new_lower, new_body, new_upper))


def _decompose_model(model: _BlockData, max_leaf_nnz: Optional[int] = None,
                     min_partition_ratio: float = 1.25, limit_num_stages: bool = True):
    """
    Parameters
    ----------
    model: _BlockData
        The model to decompose
    max_leaf_nnz: int
        maximum number nonzeros in the constraint jacobian of the leaves
    min_partition_ratio: float
        If the partition ration is less than min_partition_ratio, the partition is not
        accepted and partitioning stops. This value should be between 1 and 2.
    limit_num_stages: bool
        If True, partitioning will stop before the number of stages produced exceeds
        round(math.log10(number of nonzeros in the constraint jacobian of model))

    Returns
    -------
    new_model: TreeBlockData
        The decomposed model
    component_map: pe.ComponentMap
        A ComponentMap mapping varialbes and constraints in model to those in new_model
    termination_reason: DecompositionStatus
        An enum member from DecompositionStatus
    """

    # by reformulating the objective, we can make better use of the incumbent when
    # doing OBBT
    _reformulate_objective(model)
    # we don't want the original param objects to be in the new model
    _eliminate_mutable_params(model)

    graph = convert_pyomo_model_to_bipartite_graph(model)
    logger.debug('converted pyomo model to bipartite graph')
    original_nnz = graph.number_of_edges()
    if limit_num_stages:
        max_stages = round(math.log10(original_nnz))
    else:
        max_stages = math.inf
    logger.debug('NNZ in original graph: {0}'.format(original_nnz))
    logger.debug('maximum number of stages: {0}'.format(max_stages))
    if max_leaf_nnz is None:
        max_leaf_nnz = 0.1 * original_nnz

    if original_nnz <= max_leaf_nnz or num_cons_in_graph(graph) <= 1:
        if original_nnz <= max_leaf_nnz:
            logger.debug('too few NNZ in original graph; not decomposing')
        else:
            logger.debug('Cannot decompose graph with less than 2 constraints.')
        new_model = TreeBlock(concrete=True)
        new_model.setup(children_keys=list())
        component_map = build_pyomo_model_from_graph(graph=graph, block=new_model)
        termination_reason = DecompositionStatus.problem_too_small
        logger.debug('done building pyomo model from graph')
    else:
        root_tree, partitioning_ratio = split_metis(graph=graph, model=model)
        logger.debug('partitioned original tree; partitioning ratio: {ratio}'.format(
            ratio=partitioning_ratio))
        if partitioning_ratio < min_partition_ratio:
            logger.debug('obtained bad partitioning ratio; abandoning partition')
            new_model = TreeBlock(concrete=True)
            new_model.setup(children_keys=list())
            component_map = build_pyomo_model_from_graph(graph=graph, block=new_model)
            termination_reason = DecompositionStatus.bad_ratio
            logger.debug('done building pyomo model from graph')
        else:
            parent = root_tree

            termination_reason = DecompositionStatus.normal
            needs_split = list()
            for child in parent.children:
                logger.debug(
                    'number of NNZ in child: {0}'.format(child.number_of_edges()))
                if child.number_of_edges() > max_leaf_nnz and num_cons_in_graph(
                        child) > 1:
                    needs_split.append((child, parent, 1))

            while len(needs_split) > 0:
                logger.debug('needs_split: {0}'.format(str(needs_split)))
                _graph, _parent, _stage = needs_split.pop()
                try:
                    if _stage + 1 >= max_stages:
                        logger.debug(f'stage {_stage}: not partitiong graph with '
                                     f'{_graph.number_of_edges()} NNZ due to the max '
                                     f'stages rule;')
                        continue
                    logger.debug(f'stage {_stage}: partitioning graph with '
                                 f'{_graph.number_of_edges()} NNZ')
                    sub_tree, partitioning_ratio = split_metis(graph=_graph,
                                                               model=model)
                    logger.debug(
                        'partitioning ratio: {ratio}'.format(ratio=partitioning_ratio))
                    if partitioning_ratio > min_partition_ratio:
                        logger.debug('partitioned {0}'.format(str(_graph)))
                        _parent.children.discard(_graph)
                        _parent.children.add(sub_tree)

                        for child in sub_tree.children:
                            logger.debug('number of NNZ in child: {0}'.format(
                                child.number_of_edges()))
                            if (child.number_of_edges() > max_leaf_nnz
                                    and num_cons_in_graph(child) > 1):
                                needs_split.append((child, sub_tree, _stage + 1))
                    else:
                        logger.debug(
                            'obtained bad partitioning ratio; abandoning partition')
                except DecompositionError:
                    termination_reason = DecompositionStatus.error
                    logger.error('failed to partition graph with {0} NNZ'.format(
                        _graph.number_of_edges()))

            logger.debug('Tree Info:')
            root_tree.log()

            new_model = TreeBlock(concrete=True)
            component_map = root_tree.build_pyomo_model(block=new_model)
            logger.debug('done building pyomo model from tree')

    obj = get_objective(model)
    if obj is not None:
        var_map = {id(k): v for k, v in component_map.items()}
        new_model.objective = pe.Objective(
            expr=replace_expressions(obj.expr, substitution_map=var_map,
                                     remove_named_expressions=True),
            sense=obj.sense)
        logger.debug('done adding objective to new model')
    else:
        logger.debug('No objective was found to add to the new model')

    return new_model, component_map, termination_reason


def decompose_model(model: _BlockData, max_leaf_nnz: Optional[int] = None,
                    min_partition_ratio: float = 1.25, limit_num_stages: bool = True):
    """
    Parameters
    ----------
    model: _BlockData
        The model to decompose
    max_leaf_nnz: int
        maximum number nonzeros in the constraint jacobian of the leaves
    min_partition_ratio: float
        If the partition ration is less than min_partition_ratio, the partition is not
        accepted and partitioning stops. This value should be between 1 and 2.
    limit_num_stages: bool
        If True, partitioning will stop before the number of stages produced exceeds
        round(math.log10(number of nonzeros in the constraint jacobian of model))

    Returns
    -------
    new_model: TreeBlockData
        The decomposed model
    component_map: pe.ComponentMap
        A ComponentMap mapping varialbes and constraints in model to those in new_model
    termination_reason: DecompositionStatus
        An enum member from DecompositionStatus
    """
    # we have to clone the model because we modify it in _refine_partition
    all_comps = list(ComponentSet(
        nonrelaxation_component_data_objects(model, pe.Var, descend_into=True)))
    all_comps.extend(ComponentSet(
        nonrelaxation_component_data_objects(model, pe.Constraint, active=True,
                                             descend_into=True)))
    all_comps.extend(relaxation_data_objects(model, descend_into=True, active=True))
    all_comps.extend(ComponentSet(
        nonrelaxation_component_data_objects(model, pe.Objective, active=True,
                                             descend_into=True)))
    tmp_name = unique_component_name(model, 'all_comps')
    setattr(model, tmp_name, all_comps)
    new_model = model.clone()
    old_to_new_comps_map = pe.ComponentMap(zip(getattr(model, tmp_name),
                                               getattr(new_model, tmp_name)))
    delattr(model, tmp_name)
    delattr(new_model, tmp_name)
    model = new_model

    tmp = _decompose_model(model, max_leaf_nnz=max_leaf_nnz,
                           min_partition_ratio=min_partition_ratio,
                           limit_num_stages=limit_num_stages)
    tree_model, component_map, termination_reason = tmp

    for orig_comp, clone_comp in list(old_to_new_comps_map.items()):
        if clone_comp in component_map:
            old_to_new_comps_map[orig_comp] = component_map[clone_comp]

    return tree_model, old_to_new_comps_map, termination_reason


def collect_vars_to_tighten_from_graph(graph):
    vars_to_tighten = ComponentSet()

    for n in graph.nodes():
        if n.is_rel():
            rel: BaseRelaxationData = n.comp
            if rel.is_rhs_convex() and rel.relaxation_side == RelaxationSide.UNDER and not rel.use_linear_relaxation:
                continue
            if rel.is_rhs_concave() and rel.relaxation_side == RelaxationSide.OVER and not rel.use_linear_relaxation:
                continue
            vars_to_tighten.update(rel.get_rhs_vars())
        elif n.is_var():
            v = n.comp
            if v.is_binary() or v.is_integer():
                vars_to_tighten.add(v)

    return vars_to_tighten


def collect_vars_to_tighten(block):
    graph = convert_pyomo_model_to_bipartite_graph(block)
    vars_to_tighten = collect_vars_to_tighten_from_graph(graph=graph)
    return vars_to_tighten


def collect_vars_to_tighten_by_block(m, method):
    """
    Parameters
    ----------
    m: TreeBlockData
    method: str
        'full_space', 'dbt', or 'leaves'

    Returns
    -------
    vars_to_tighten_by_block: dict
        maps Block to ComponentSet of Var
    """
    assert method in {'full_space', 'dbt', 'leaves'}

    vars_to_tighten_by_block = dict()

    assert isinstance(m, TreeBlockData)

    all_vars_to_account_for = collect_vars_to_tighten(m)

    for stage in range(m.num_stages()):
        for block in m.stage_blocks(stage, active=True):
            if block.is_leaf():
                vars_to_tighten_by_block[block] = collect_vars_to_tighten(block=block)
            elif method == 'leaves':
                vars_to_tighten_by_block[block] = ComponentSet()
            elif method == 'full_space':
                vars_to_tighten_by_block[block] = ComponentSet()
            else:
                vars_to_tighten_by_block[block] = ComponentSet()
                for c in block.linking_constraints.values():
                    if c.active:
                        vars_in_con = list(identify_variables(c.body))
                        vars_to_tighten_by_block[block].add(vars_in_con[0])

    for block, vars_to_tighten in vars_to_tighten_by_block.items():
        for v in vars_to_tighten:
            all_vars_to_account_for.discard(v)

    if len(all_vars_to_account_for) != 0:
        raise RuntimeError('There are variables that need tightened that are unaccounted for!')

    return vars_to_tighten_by_block


class OBBTMethod(enum.Enum):
    FULL_SPACE = 1
    DECOMPOSED = 2
    LEAVES = 3


class FilterMethod(enum.Enum):
    NONE = 1
    AGGRESSIVE = 2


class DBTInfo(object):
    """
    Attributes
    ----------
    num_coupling_vars_to_tighten: int
        The total number of coupling variables that need tightened. Note that this includes
        coupling variables that get filtered. If you subtract num_coupling_vars_attempted 
        and num_coupling_vars_filtered from num_coupling_vars_to_tighten, you should get 
        the number of coupling variables that were not tightened due to a time limit.
    num_coupling_vars_attempted: int
        The number of coupling variables for which tightening was attempted.
    num_coupling_vars_successful: int
        The number of coupling variables for which tightening was attempted and the solver 
        terminated optimally.
    num_coupling_vars_filtered: int
        The number of coupling vars that did not need to be tightened (identified by filtering).
    num_vars_to_tighten: int
        The total number of nonlinear and discrete variables that need tightened. Note that 
        this includes variables that get filtered. If you subtract num_vars_attempted and 
        num_vars_filtered from num_vars_to_tighten, you should get the number of nonlinear 
        and discrete variables that were not tightened due to a time limit.
    num_vars_attempted: int
        The number of variables for which tightening was attempted.
    num_vars_successful: int
        The number of variables for which tightening was attempted and the solver 
        terminated optimally.
    num_vars_filtered: int
        The number of vars that did not need to be tightened (identified by filtering).
    """
    def __init__(self):
        self.num_coupling_vars_to_tighten = None
        self.num_coupling_vars_attempted = None
        self.num_coupling_vars_successful = None
        self.num_coupling_vars_filtered = None
        self.num_vars_to_tighten = None
        self.num_vars_attempted = None
        self.num_vars_successful = None
        self.num_vars_filtered = None

    def __str__(self):
        s = f'num_coupling_vars_to_tighten: {self.num_coupling_vars_to_tighten}\n'
        s += f'num_coupling_vars_attempted: {self.num_coupling_vars_attempted}\n'
        s += f'num_coupling_vars_successful: {self.num_coupling_vars_successful}\n'
        s += f'num_coupling_vars_filtered: {self.num_coupling_vars_filtered}\n'
        s += f'num_vars_to_tighten: {self.num_vars_to_tighten}\n'
        s += f'num_vars_attempted: {self.num_vars_attempted}\n'
        s += f'num_vars_successful: {self.num_vars_successful}\n'
        s += f'num_vars_filtered: {self.num_vars_filtered}\n'
        return s


def _update_var_bounds(varlist, new_lower_bounds, new_upper_bounds, feasibility_tol, safety_tol, max_acceptable_bound):
    for ndx, v in enumerate(varlist):
        new_lb = new_lower_bounds[ndx]
        new_ub = new_upper_bounds[ndx]
        orig_lb = v.lb
        orig_ub = v.ub

        if new_lb is None:
            new_lb = -math.inf
        if new_ub is None:
            new_ub = math.inf
        if orig_lb is None:
            orig_lb = -math.inf
        if orig_ub is None:
            orig_ub = math.inf

        rel_lb_safety = safety_tol * abs(new_lb)
        rel_ub_safety = safety_tol * abs(new_ub)
        new_lb -= max(safety_tol, rel_lb_safety)
        new_ub += max(safety_tol, rel_ub_safety)

        if new_lb < -max_acceptable_bound:
            new_lb = -math.inf
        if new_ub > max_acceptable_bound:
            new_ub = math.inf

        if new_lb > new_ub:
            msg = 'variable ub is less than lb; var: {0}; lb: {1}; ub: {2}'.format(str(v), new_lb, new_ub)
            if new_lb > new_ub + feasibility_tol:
                raise ValueError(msg)
            else:
                logger.warning(msg + '; decreasing lb and increasing ub by {0}'.format(feasibility_tol))
                warnings.warn(msg)
                new_lb -= feasibility_tol
                new_ub += feasibility_tol

        if new_lb < orig_lb:
            new_lb = orig_lb
        if new_ub > orig_ub:
            new_ub = orig_ub

        if new_lb > -math.inf:
            v.setlb(new_lb)
        if new_ub < math.inf:
            v.setub(new_ub)


def perform_dbt(relaxation, solver, obbt_method=OBBTMethod.DECOMPOSED,
                filter_method=FilterMethod.AGGRESSIVE, time_limit=math.inf,
                objective_bound=None, with_progress_bar=False, parallel=False,
                vars_to_tighten_by_block=None, feasibility_tol=0,
                safety_tol=0, max_acceptable_bound=math.inf, update_relaxations_between_stages=True):
    """This function performs optimization-based bounds tightening (OBBT) with a decomposition scheme.

    Parameters
    ----------
    relaxation: dbt.decomp.decompose.TreeBlockData
        The relaxation to use for OBBT.
    solver: pyomo solver object
        The solver to use for the OBBT problems.
    obbt_method: OBBTMethod
        An enum member from OBBTMethod. The default is OBBTMethod.DECOMPOSED. If obbt_method
        is OBBTMethod.DECOMPOSED, then only the coupling variables in the linking constraints
        will be tightened with non-leaf blocks in relaxation. The nonlinear and discrete
        variables will only be tightened with the leaf blocks in relaxation. See the
        documentation on TreeBlockData for more details on leaf blocks. If the method is
        OBBTMethod.FULL_SPACE, then all of the nonlinear and discrete variables will
        be tightened with the root block from relaxation. If the method is
        OBBTMethod.LEAVES, then the nonlinear and discrete variables will be tightened
        with the leaf blocks from relaxation (none of the coupling variables will be
        tightened).
    filter_method: FilterMethod
        An enum member from FilterMethod. The default is FilterMethod.AGGRESSIVE. If
        filter_method is FilterMethod.AGGRESSIVE, then aggressive filtering will be
        performed at every stage of OBBT using the
        coramin.domain_reduction.filters.aggressive_filter function which is based on

            Gleixner, Ambros M., et al. "Three enhancements for
            optimization-based bound tightening." Journal of Global
            Optimization 67.4 (2017): 731-757.

        If filter_method is FilterMethod.NONE, then no filtering will be performed.
    time_limit: float
        If the time spent in this function exceeds time_limit, OBBT will be terminated
        early.
    objective_bound: float
        A lower or upper bound on the objective. If this is not None, then a constraint will be added to the
        bounds tightening problems constraining the objective to be less than/greater than objective_bound.
    with_progress_bar: bool
    parallel: bool
        If True, then OBBT will automatically be performed in parallel if mpirun or mpiexec was used;
        If False, then OBBT will not run in parallel even if mpirun or mpiexec was used;
    vars_to_tighten_by_block: dict
        Dictionary mapping TreeBlockData to ComponentSet. This dictionary indicates which variables
        should be tightened with which parts of the TreeBlockData. If None is passed (default=None),
        then, the function collect_vars_to_tighten_by_block is used to get the dict.
    feasibility_tol: float
        If the lower bound for a computed variable is larger than the computed upper bound by more than
        feasibility_tol, then an error is raised. If the computed lower bound is larger than the computed
        upper bound, but by less than feasibility_tol, then the computed lower bound is decreased by
        feasibility tol (but will not be set lower than the original lower bound) and the computed upper
        bound is increased by feasibility_tol (but will not be set higher than the original upper bound).
    safety_tol: float
        Computed lower bounds will be decreased by max(safety_tol, safety_tol*abs(new_lb) and
        computed upper bounds will be increased by max(safety_tol, safety_tol*abs(new_ub) where
        new_lb and new_ub are the bounds computed from OBBT/DBT. The purpose of this is to
        account for numerical error in the solution of the OBBT problems and to avoid cutting
        off valid portions of the feasible region.
    max_acceptable_bound: float
        If the upper bound computed for a variable is larger than max_acceptable_bound, then the 
        computed bound will be rejected. If the lower bound computed for a variable is less than 
        -max_acceptable_bound, then the computed bound will be rejected.
    update_relaxations_between_stages: bool
        This is meant for unit testing only and should not be modified

    Returns
    -------
    dbt_info: DBTInfo

    """
    t0 = time.time()
    
    if not isinstance(relaxation, TreeBlockData):
        raise ValueError('relaxation must be an instance of dbt.decomp.TreeBlockData.')
    if obbt_method not in OBBTMethod:
        raise ValueError('obbt_method must be a member of OBBTMethod.')
    if filter_method not in FilterMethod:
        raise ValueError('filter_method must a member of FilterMethod.')
    if isinstance(solver, PersistentSolver):
        using_persistent_solver = True
    else:
        using_persistent_solver = False

    dbt_info = DBTInfo()
    dbt_info.num_coupling_vars_to_tighten = 0
    dbt_info.num_coupling_vars_attempted = 0
    dbt_info.num_coupling_vars_successful = 0
    dbt_info.num_coupling_vars_filtered = 0
    dbt_info.num_vars_to_tighten = 0
    dbt_info.num_vars_attempted = 0
    dbt_info.num_vars_successful = 0
    dbt_info.num_vars_filtered = 0

    assert obbt_method in OBBTMethod
    if vars_to_tighten_by_block is None:
        if obbt_method == OBBTMethod.DECOMPOSED:
            _method = 'dbt'
        elif obbt_method == OBBTMethod.FULL_SPACE:
            _method = 'full_space'
        else:
            _method = 'leaves'
        vars_to_tighten_by_block = collect_vars_to_tighten_by_block(relaxation, _method)

    var_to_relaxation_map = pe.ComponentMap()
    for r in relaxation_data_objects(relaxation, descend_into=True, active=True):
        for v in r.get_rhs_vars():
            if v not in var_to_relaxation_map:
                var_to_relaxation_map[v] = list()
            var_to_relaxation_map[v].append(r)

    num_stages = relaxation.num_stages()

    for stage in range(num_stages):
        stage_blocks = list(relaxation.stage_blocks(stage))
        for block in stage_blocks:
            vars_to_tighten = vars_to_tighten_by_block[block]
            if obbt_method == OBBTMethod.FULL_SPACE or block.is_leaf():
                dbt_info.num_vars_to_tighten += 2 * len(vars_to_tighten)
            else:
                dbt_info.num_coupling_vars_to_tighten += 2 * len(vars_to_tighten)

    if obbt_method == OBBTMethod.FULL_SPACE:
        all_vars_to_tighten = ComponentSet()
        for block, block_vars_to_tighten in vars_to_tighten_by_block.items():
            all_vars_to_tighten.update(block_vars_to_tighten)
        if filter_method == FilterMethod.AGGRESSIVE:
            logger.debug('starting full space filter')
            res = aggressive_filter(candidate_variables=all_vars_to_tighten, relaxation=relaxation,
                                    solver=solver, tolerance=1e-4, objective_bound=objective_bound)
            full_space_lb_vars, full_space_ub_vars = res
            logger.debug('finished full space filter')
        else:
            full_space_lb_vars = all_vars_to_tighten
            full_space_ub_vars = all_vars_to_tighten
    else:
        full_space_lb_vars = None
        full_space_ub_vars = None
    
    for stage in range(num_stages):
        logger.info(f'Performing DBT on stage {stage+1} of {num_stages}')
        if time.time() - t0 >= time_limit:
            break
        
        stage_blocks = list(relaxation.stage_blocks(stage))
        logger.debug('DBT stage {0} of {1} with {1} blocks'.format(stage, num_stages, len(stage_blocks)))

        for block_ndx, block in enumerate(stage_blocks):
            logger.info(f'performing DBT on block {block_ndx+1} of {len(stage_blocks)} in stage {stage+1}')
            if time.time() - t0 >= time_limit:
                break

            if obbt_method in {OBBTMethod.LEAVES, OBBTMethod.FULL_SPACE} and (not block.is_leaf()):
                continue
            if obbt_method == OBBTMethod.FULL_SPACE:
                block_to_tighten_with = relaxation
                _ub = objective_bound
            else:
                block_to_tighten_with = block
                if stage == 0:
                    _ub = objective_bound
                else:
                    _ub = None
                    
            vars_to_tighten = vars_to_tighten_by_block[block]

            if filter_method == FilterMethod.AGGRESSIVE:
                logger.debug('starting filter')
                if obbt_method == OBBTMethod.FULL_SPACE:
                    lb_vars = ComponentSet([v for v in vars_to_tighten if v in full_space_lb_vars])
                    ub_vars = ComponentSet([v for v in vars_to_tighten if v in full_space_ub_vars])
                else:
                    res = aggressive_filter(candidate_variables=vars_to_tighten, relaxation=block_to_tighten_with,
                                            solver=solver, tolerance=1e-4, objective_bound=_ub)
                    lb_vars, ub_vars = res
                if block.is_leaf():
                    dbt_info.num_vars_filtered += 2*len(vars_to_tighten) - len(lb_vars) - len(ub_vars)
                else:
                    dbt_info.num_coupling_vars_filtered += 2*len(vars_to_tighten) - len(lb_vars) - len(ub_vars)
                logger.debug('done filtering')
            else:
                lb_vars = list(vars_to_tighten)
                ub_vars = list(vars_to_tighten)

            logger.debug(f'performing OBBT (LB) on variables {str([str(i) for i in lb_vars])}')
            res = normal_obbt(block_to_tighten_with, solver=solver, varlist=lb_vars,
                              objective_bound=_ub, with_progress_bar=with_progress_bar,
                              direction='lbs', time_limit=(time_limit - (time.time() - t0)),
                              update_bounds=False, parallel=parallel, collect_obbt_info=True,
                              progress_bar_string=f'DBT LBs Stage {stage+1} of {num_stages} Block {block_ndx+1} of {len(stage_blocks)}')
            lower, unused_upper, obbt_info = res
            if block.is_leaf():
                dbt_info.num_vars_attempted += obbt_info.num_problems_attempted
                dbt_info.num_vars_successful += obbt_info.num_successful_problems
            else:
                dbt_info.num_coupling_vars_attempted += obbt_info.num_problems_attempted
                dbt_info.num_coupling_vars_successful += obbt_info.num_successful_problems

            logger.debug('done tightening lbs')

            logger.debug(f'performing OBBT (UB) on variables {str([str(i) for i in ub_vars])}')
            res = normal_obbt(block_to_tighten_with, solver=solver, varlist=ub_vars,
                              objective_bound=_ub, with_progress_bar=with_progress_bar,
                              direction='ubs', time_limit=(time_limit - (time.time() - t0)),
                              update_bounds=False, parallel=parallel, collect_obbt_info=True,
                              progress_bar_string=f'DBT UBs Stage {stage+1} of {num_stages} Block {block_ndx+1} of {len(stage_blocks)}')

            unused_lower, upper, obbt_info = res
            if block.is_leaf():
                dbt_info.num_vars_attempted += obbt_info.num_problems_attempted
                dbt_info.num_vars_successful += obbt_info.num_successful_problems
            else:
                dbt_info.num_coupling_vars_attempted += obbt_info.num_problems_attempted
                dbt_info.num_coupling_vars_successful += obbt_info.num_successful_problems

            _update_var_bounds(varlist=lb_vars, new_lower_bounds=lower,
                               new_upper_bounds=unused_upper, feasibility_tol=feasibility_tol,
                               safety_tol=safety_tol, max_acceptable_bound=max_acceptable_bound)

            _update_var_bounds(varlist=ub_vars, new_lower_bounds=unused_lower,
                               new_upper_bounds=upper, feasibility_tol=feasibility_tol,
                               safety_tol=safety_tol, max_acceptable_bound=max_acceptable_bound)

            if update_relaxations_between_stages:
                # this is needed to ensure consistency for parallel computing; this accounts
                # for side effects from the OBBT problems; in particular, if the solver ever
                # rebuilds relaxations, then the processes could become out of sync without
                # this code
                all_tightened_vars = ComponentSet(lb_vars)
                all_tightened_vars.update(ub_vars)
                for v in all_tightened_vars:
                    if v in var_to_relaxation_map:
                        for r in var_to_relaxation_map[v]:
                            r.rebuild()

            logger.debug('done tightening ubs')

            if not block.is_leaf():
                for c in block.linking_constraints.values():
                    fbbt(c)

    return dbt_info


def push_integers(block):
    """
    Parameters
    ----------
    block: pyomo.core.base.block._BlockData
        The block for which integer variables should be relaxed.

    Returns
    -------
    relaxed_binary_vars: ComponentSet of pyomo.core.base.var._GeneralVarData
    relaxed_integer_vars: ComponentSet or pyomo.core.base.var._GeneralVarData
    """
    relaxed_binary_vars = ComponentSet()
    relaxed_integer_vars = ComponentSet()
    for v in block.component_data_objects(pe.Var, descend_into=True, sort=True):
        if v.fixed:
            continue
        if v.is_binary():
            relaxed_binary_vars.add(v)
            orig_lb = v.lb
            orig_ub = v.ub
            v.domain = pe.Reals
            v.setlb(orig_lb)
            v.setub(orig_ub)
        elif v.is_integer():
            relaxed_integer_vars.add(v)
            v.domain = pe.Reals

    return relaxed_binary_vars, relaxed_integer_vars


def pop_integers(relaxed_binary_vars, relaxed_integer_vars):
    for v in relaxed_binary_vars:
        v.domain = pe.Binary
    for v in relaxed_integer_vars:
        v.domain = pe.Integers


def perform_dbt_with_integers_relaxed(relaxation, solver, obbt_method=OBBTMethod.DECOMPOSED,
                                      filter_method=FilterMethod.AGGRESSIVE, time_limit=math.inf,
                                      objective_bound=None, with_progress_bar=False, parallel=False,
                                      vars_to_tighten_by_block=None, feasibility_tol=0,
                                      integer_tol=1e-2, safety_tol=0, max_acceptable_bound=math.inf):
    """
    This function performs optimization-based bounds tightening (OBBT) with a decomposition scheme.
    However, all OBBT problems are solved with the binary and integer variables relaxed.

    Parameters
    ----------
    relaxation: dbt.decomp.decompose.TreeBlockData
        The relaxation to use for OBBT.
    solver: pyomo solver object
        The solver to use for the OBBT problems.
    obbt_method: OBBTMethod
        An enum member from OBBTMethod. The default is OBBTMethod.DECOMPOSED. If obbt_method
        is OBBTMethod.DECOMPOSED, then only the coupling variables in the linking constraints
        will be tightened with non-leaf blocks in relaxation. The nonlinear and discrete
        variables will only be tightened with the leaf blocks in relaxation. See the
        documentation on TreeBlockData for more details on leaf blocks. If the method is
        OBBTMethod.FULL_SPACE, then all of the nonlinear and discrete variables will
        be tightened with the root block from relaxation. If the method is
        OBBTMethod.LEAVES, then the nonlinear and discrete variables will be tightened
        with the leaf blocks from relaxation (none of the coupling variables will be
        tightened).
    filter_method: FilterMethod
        An enum member from FilterMethod. The default is FilterMethod.AGGRESSIVE. If
        filter_method is FilterMethod.AGGRESSIVE, then aggressive filtering will be
        performed at every stage of OBBT using the
        coramin.domain_reduction.filters.aggressive_filter function which is based on

            Gleixner, Ambros M., et al. "Three enhancements for
            optimization-based bound tightening." Journal of Global
            Optimization 67.4 (2017): 731-757.

        If filter_method is FilterMethod.NONE, then no filtering will be performed.
    time_limit: float
        If the time spent in this function exceeds time_limit, OBBT will be terminated
        early.
    objective_bound: float
        A lower or upper bound on the objective. If this is not None, then a constraint will be added to the
        bounds tightening problems constraining the objective to be less than/greater than objective_bound.
    with_progress_bar: bool
    parallel: bool
        If True, then OBBT will automatically be performed in parallel if mpirun or mpiexec was used;
        If False, then OBBT will not run in parallel even if mpirun or mpiexec was used;
    vars_to_tighten_by_block: dict
        Dictionary mapping TreeBlockData to ComponentSet. This dictionary indicates which variables
        should be tightened with which parts of the TreeBlockData. If None is passed (default=None),
        then, the function collect_vars_to_tighten_by_block is used to get the dict.
    feasibility_tol: float
        If the lower bound computed for a variable is larger than the computed upper bound by more than
        feasibility_tol, then an error is raised. If the computed lower bound is larger than the computed
        upper bound, but by less than feasibility_tol, then the computed lower bound is decreased by
        feasibility tol (but will not be set lower than the original lower bound) and the computed upper
        bound is increased by feasibility_tol (but will not be set higher than the original upper bound).
    integer_tol: float
        If the lower bound computed for an integer variable is greater than the largest integer less than
        the computed lower bound by more than integer_tol, then the lower bound is increased to the smallest
        integer greater than the computed lower bound. Similar logic holds for the upper bound.
    safety_tol: float
        Computed lower bounds will be decreased by max(safety_tol, safety_tol*abs(new_lb) and
        computed upper bounds will be increased by max(safety_tol, safety_tol*abs(new_ub) where
        new_lb and new_ub are the bounds computed from OBBT/DBT. The purpose of this is to
        account for numerical error in the solution of the OBBT problems and to avoid cutting
        off valid portions of the feasible region.
    max_acceptable_bound: float
        If the upper bound computed for a variable is larger than max_acceptable_bound, then the 
        computed bound will be rejected. If the lower bound computed for a variable is less than 
        -max_acceptable_bound, then the computed bound will be rejected.

    Returns
    -------
    dbt_info: DBTInfo
    """
    assert obbt_method in OBBTMethod
    if vars_to_tighten_by_block is None:
        if obbt_method == OBBTMethod.DECOMPOSED:
            _method = 'dbt'
        elif obbt_method == OBBTMethod.FULL_SPACE:
            _method = 'full_space'
        else:
            _method = 'leaves'
        vars_to_tighten_by_block = collect_vars_to_tighten_by_block(relaxation, _method)

    relaxed_binary_vars, relaxed_integer_vars = push_integers(relaxation)

    dbt_info = perform_dbt(relaxation=relaxation,
                           solver=solver,
                           obbt_method=obbt_method,
                           filter_method=filter_method,
                           time_limit=time_limit,
                           objective_bound=objective_bound,
                           with_progress_bar=with_progress_bar,
                           parallel=parallel,
                           vars_to_tighten_by_block=vars_to_tighten_by_block,
                           feasibility_tol=feasibility_tol,
                           safety_tol=safety_tol,
                           max_acceptable_bound=max_acceptable_bound)

    pop_integers(relaxed_binary_vars, relaxed_integer_vars)

    for v in (list(relaxed_binary_vars) + list(relaxed_integer_vars)):
        lb = v.lb
        ub = v.ub
        if lb is None:
            lb = -math.inf
        if ub is None:
            ub = math.inf
        if lb > -math.inf:
            lb = max(math.floor(lb), math.ceil(lb - integer_tol))
        if ub < math.inf:
            ub = min(math.ceil(ub), math.floor(ub + integer_tol))
        if lb > -math.inf:
            v.setlb(lb)
        if ub < math.inf:
            v.setub(ub)

    return dbt_info



from enum import IntEnum, Enum


class EigenValueBounder(IntEnum):
    Gershgorin = 1
    GershgorinWithSimplification = 2
    LinearProgram = 3
    Global = 4


class Effort(IntEnum):
    none = 0
    very_low = 1
    low = 2
    medium = 3
    high = 4
    very_high = 5


class RelaxationSide(IntEnum):
    UNDER = 1
    OVER = 2
    BOTH = 3


class FunctionShape(IntEnum):
    LINEAR = 1
    CONVEX = 2
    CONCAVE = 3
    UNKNOWN = 4



from .relaxations_base import BaseRelaxation, BaseRelaxationData, BasePWRelaxation, BasePWRelaxationData
from .mccormick import PWMcCormickRelaxation, PWMcCormickRelaxationData
from .segments import compute_k_segment_points
from .univariate import PWXSquaredRelaxation, PWXSquaredRelaxationData
from .univariate import PWUnivariateRelaxation, PWUnivariateRelaxationData
from .univariate import PWArctanRelaxation, PWArctanRelaxationData
from .univariate import PWSinRelaxation, PWSinRelaxationData
from .univariate import PWCosRelaxation, PWCosRelaxationData
from .auto_relax import relax
from .alphabb import AlphaBBRelaxationData, AlphaBBRelaxation
from .multivariate import MultivariateRelaxationData, MultivariateRelaxation
from .iterators import relaxation_data_objects, nonrelaxation_component_data_objects



from coramin.utils.coramin_enums import EigenValueBounder, RelaxationSide
from coramin.relaxations.custom_block import declare_custom_block
from coramin.relaxations.relaxations_base import BaseRelaxationData, ComponentWeakRef
from coramin.relaxations.hessian import Hessian
from typing import Optional, Tuple
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.expr.numeric_expr import ExpressionBase
from pyomo.core.expr.visitor import identify_variables
from pyomo.contrib import appsi
from pyomo.core.base.param import ScalarParam, IndexedParam
from pyomo.core.base.set import OrderedScalarSet


@declare_custom_block(name='AlphaBBRelaxation')
class AlphaBBRelaxationData(BaseRelaxationData):
    def __init__(self, component):
        super().__init__(component)
        self._xs: Optional[Tuple[_GeneralVarData]] = None
        self._aux_var_ref = ComponentWeakRef(None)
        self._f_x_expr: Optional[ExpressionBase] = None
        self._alphabb_rhs: Optional[ExpressionBase] = None
        self._hessian: Optional[Hessian] = None
        self._alpha: Optional[ScalarParam] = None
        self._var_set: Optional[OrderedScalarSet] = None
        self._lb_params: Optional[IndexedParam] = None
        self._ub_params: Optional[IndexedParam] = None

    @property
    def hessian(self):
        return self._hessian

    @property
    def _aux_var(self):
        return self._aux_var_ref.get_component()

    def get_rhs_vars(self) -> Tuple[_GeneralVarData, ...]:
        return self._xs

    def get_rhs_expr(self) -> ExpressionBase:
        return self._f_x_expr

    def vars_with_bounds_in_relaxation(self):
        if self.is_rhs_convex():
            return list()
        else:
            return list(self._xs)

    def is_rhs_convex(self):
        return self._hessian.get_minimum_eigenvalue() >= 0

    def is_rhs_concave(self):
        return self._hessian.get_maximum_eigenvalue() <= 0

    def has_convex_underestimator(self):
        return self.relaxation_side == RelaxationSide.UNDER

    def has_concave_overestimator(self):
        return self.relaxation_side == RelaxationSide.OVER

    def _get_expr_for_oa(self):
        return self._alphabb_rhs

    def set_input(
        self,
        aux_var: _GeneralVarData,
        f_x_expr: ExpressionBase,
        relaxation_side: RelaxationSide,
        use_linear_relaxation: bool = True,
        large_coef: float = 1e5,
        small_coef: float = 1e-10,
        safety_tol: float = 1e-10,
        eigenvalue_bounder: EigenValueBounder = EigenValueBounder.LinearProgram,
        eigenvalue_opt: Optional[appsi.base.Solver] = None,
        hessian: Optional[Hessian] = None,
    ):
        del self._alpha, self._alphabb_rhs, self._var_set, self._lb_params
        del self._ub_params
        self._alpha, self._alphabb_rhs, self._var_set = None, None, None
        self._lb_params, self._ub_params = None, None
        self._relaxation_side = relaxation_side
        super().set_input(
            relaxation_side=relaxation_side,
            use_linear_relaxation=use_linear_relaxation,
            large_coef=large_coef,
            small_coef=small_coef,
            safety_tol=safety_tol
        )
        self._xs = tuple(identify_variables(f_x_expr, include_fixed=False))
        self._aux_var_ref.set_component(aux_var)
        self._f_x_expr = f_x_expr
        if hessian is None:
            hessian = Hessian(
                expr=f_x_expr, opt=eigenvalue_opt, method=eigenvalue_bounder
            )
        self._hessian = hessian

    def build(
        self,
        aux_var: _GeneralVarData,
        f_x_expr: ExpressionBase,
        relaxation_side: RelaxationSide,
        use_linear_relaxation: bool = True,
        large_coef: float = 1e5,
        small_coef: float = 1e-10,
        safety_tol: float = 1e-10,
        eigenvalue_bounder: EigenValueBounder = EigenValueBounder.LinearProgram,
        eigenvalue_opt: appsi.base.Solver = None,
    ):
        self.set_input(
            aux_var=aux_var,
            f_x_expr=f_x_expr,
            relaxation_side=relaxation_side,
            use_linear_relaxation=use_linear_relaxation,
            large_coef=large_coef,
            small_coef=small_coef,
            safety_tol=safety_tol,
            eigenvalue_bounder=eigenvalue_bounder,
            eigenvalue_opt=eigenvalue_opt,
        )
        self.rebuild()

    @property
    def use_linear_relaxation(self):
        return self._use_linear_relaxation

    @use_linear_relaxation.setter
    def use_linear_relaxation(self, value):
        self._use_linear_relaxation = value

    @property
    def relaxation_side(self):
        return BaseRelaxationData.relaxation_side.fget(self)

    @relaxation_side.setter
    def relaxation_side(self, val):
        if val != self.relaxation_side:
            raise ValueError('Cannot change the relaxation side of an AlphaBBRelaxation')
        if val == RelaxationSide.BOTH:
            raise ValueError('AlphaBBRelaxation only supports relaxation sides of UNDER or OVER, not BOTH.')

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        if self.relaxation_side == RelaxationSide.UNDER:
            alpha = max(0, -0.5 * self._hessian.get_minimum_eigenvalue())
        else:
            alpha = max(0, 0.5*self._hessian.get_maximum_eigenvalue())
            alpha = -alpha
        if self._alpha is None:
            del self._alpha, self._alphabb_rhs, self._var_set
            del self._lb_params, self._ub_params
            self._alpha = ScalarParam(mutable=True)
            self._var_set = OrderedScalarSet(initialize=list(range(len(self._xs))))
            self._lb_params = IndexedParam(self._var_set, mutable=True)
            self._ub_params = IndexedParam(self._var_set, mutable=True)
            alpha_sum = 0
            for ndx, v in enumerate(self._xs):
                p_lb = self._lb_params[ndx]
                p_ub = self._ub_params[ndx]
                alpha_sum += (v - p_lb) * (v - p_ub)
            self._alphabb_rhs = self.get_rhs_expr() + self._alpha * alpha_sum
        self._alpha.value = alpha
        for ndx, v in enumerate(self._xs):
            self._lb_params[ndx].value = v.lb
            self._ub_params[ndx].value = v.ub

        super().rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                        ensure_oa_at_vertices=ensure_oa_at_vertices)



import logging
import pyomo.environ as pe
import warnings
import math

logger = logging.getLogger(__name__)
pyo = pe


def _get_bnds_tuple(v):
    lb = pe.value(v.lb)
    ub = pe.value(v.ub)
    if lb is None:
        lb = -math.inf
    if ub is None:
        ub = math.inf

    return lb, ub


def _get_bnds_list(v):
    return list(_get_bnds_tuple(v))


def var_info_str(v):
    s = '\tVar: {0}\n'.format(v)
    return s


def bnds_info_str(vlb, vub):
    s = '\tLB: {0}\n'.format(vlb)
    s += '\tUB: {0}\n'.format(vub)
    return s


def x_pts_info_str(_x_pts):
    s = '\tx_pts: {0}\n'.format(_x_pts)
    return s


def check_var_pts(x, x_pts=None):
    xlb = pe.value(x.lb)
    xub = pe.value(x.ub)

    if xlb is None:
        xlb = -math.inf
    if xub is None:
        xub = math.inf

    raise_error = False
    raise_warning = False
    msg = None

    if xub < xlb:
        msg = 'Lower bound is larger than upper bound:\n' + var_info_str(x) + bnds_info_str(xlb, xub)
        raise_error = True

    if x_pts is not None:
        ordered = all(x_pts[i] <= x_pts[i+1] for i in range(len(x_pts)-1))
        if not ordered:
            msg = 'x_pts must be ordered:\n' + var_info_str(x) + bnds_info_str(xlb, xub) + x_pts_info_str(x_pts)
            raise_error = True

        if xlb != x_pts[0] or xub != x_pts[-1]:
            msg = ('end points of the x_pts list must be equal to the bounds on the x variable:\n' + var_info_str(x) +
                   bnds_info_str(xlb, xub) + x_pts_info_str(x_pts))
            raise_error = True

    if raise_error:
        logger.error(msg)
        raise ValueError(msg)

    if raise_warning:
        logger.warning(msg)
        warnings.warn(msg)



import sys
from pyomo.core.base.block import Block
from pyomo.core.base.indexed_component import UnindexedComponent_set
# ToDo: documentation
# ToDo: passing of kwargs down to the data object
# ToDo: figure out if the setattr's are necessary in the decorator
# ToDo: decide if we need the decorator (it actually does not do much
#       and can be replaced by one more class declaration that might be "easier"
#       for the user anyway?
'''
This module implements meta classes and a decorator to make
it easier to create derived block types. With the decorator,
you only need to inherit from _BlockData.

# ToDo: Document this custom block code with an example
'''

class _IndexedCustomBlockMeta(type):
    """Metaclass for creating an indexed block with
    a custom block data type."""

    def __new__(meta, name, bases, dct):
        def __init__(self, *args, **kwargs):
            bases[0].__init__(self, *args, **kwargs)
        dct["__init__"] = __init__
        return type.__new__(meta, name, bases, dct)

class _ScalarCustomBlockMeta(type):
    '''Metaclass used to create a scalar block with a
    custom block data type
    '''
    def __new__(meta, name, bases, dct):
        def __init__(self, *args, **kwargs):
            # bases[0] is the custom block data object
            bases[0].__init__(self, component=self)
            # bases[1] is the custom block object that
            # is used for declaration
            bases[1].__init__(self, *args, **kwargs)
        dct["__init__"] = __init__
        return type.__new__(meta, name, bases, dct)

class CustomBlock(Block):
    ''' This CustomBlock is the base class that allows
    for easy creation of specialized derived blocks
    '''
    def __new__(cls, *args, **kwds):
        if cls.__name__.startswith('_Indexed') or \
                cls.__name__.startswith('_Scalar'):
            # we are entering here the second time (recursive)
            # therefore, we need to create what we have
            return super(CustomBlock, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args)==1):
            bname = "_Scalar{}".format(cls.__name__)
            n = _ScalarCustomBlockMeta(bname, (cls._ComponentDataClass, cls),{})
            return n.__new__(n)
        else:
            bname = "_Indexed{}".format(cls.__name__)
            n = _IndexedCustomBlockMeta(bname, (cls,), {})
            return n.__new__(n)


def declare_custom_block(name):
    ''' Decorator to declare the custom component
    that goes along with a custom block data

    @declare_custom_block(name=FooBlock)
    class FooBlockData(_BlockData):
       # custom block data class
    '''
    def proc_dec(cls):
        # this is the decorator function that
        # creates the block component class
        c = type(
            name, # name of new class
            (CustomBlock,), # base classes
            {"__module__": cls.__module__, "_ComponentDataClass": cls}) # magic to fix the module

        # are these necessary?
        setattr(sys.modules[cls.__module__], name, c)
        setattr(cls, '_orig_name', name)
        setattr(cls, '_orig_module', cls.__module__)
        return cls
    return proc_dec



import pyomo.environ as pyo
import warnings
import logging

logger = logging.getLogger(__name__)


def compute_k_segment_points(v, k):
    """
    Return a list of points that generats k segments between v.lb and v.ub

    Parameters
    ----------
    v: pyo.Var
    k: int

    Returns
    -------
    pts: list of float
    """
    delta = (pyo.value(v.ub) - pyo.value(v.lb)) / k
    pts = [pyo.value(v.lb) + i * delta for i in range(k)]
    pts.append(pyo.value(v.ub))
    return pts



import math

import coramin
from coramin.third_party.minlplib_tools import get_minlplib, get_minlplib_instancedata
import unittest
from pyomo.contrib import appsi
import os
import logging
from suspect.pyomo.osil_reader import read_osil
import math
from pyomo.common import download
import pyomo.environ as pe
from pyomo.core.base.block import _BlockData


def _get_sol(pname):
    current_dir = os.getcwd()
    target_fname = os.path.join(current_dir, f'{pname}.sol')
    downloader = download.FileDownloader()
    downloader.set_destination_filename(target_fname)
    downloader.get_binary_file(f'http://www.minlplib.org/sol/{pname}.p1.sol')
    res = dict()
    f = open(target_fname, 'r')
    for line in f.readlines():
        l = line.split()
        vname = l[0]
        vval = float(l[1])
        if vname != 'objvar':
            res[vname] = vval
    f.close()
    return res


class Helper(unittest.TestCase):
    def _check_relative_diff(self, expected, got, abs_tol=1e-3, rel_tol=1e-3):
        abs_diff = abs(expected - got)
        if expected == 0:
            rel_diff = math.inf
        else:
            rel_diff = abs_diff / abs(expected)
        success = abs_diff <= abs_tol or rel_diff <= rel_tol
        self.assertTrue(success, msg=f'\n    expected: {expected}\n    got: {got}\n    abs diff: {abs_diff}\n    rel diff: {rel_diff}')


class TestMultiTreeWithMINLPLib(Helper):
    @classmethod
    def setUpClass(self) -> None:
        self.test_problems = {'batch': 285506.5082,
                              'ball_mk3_10': None,
                              'ball_mk2_10': 0,
                              'syn05m': 837.73240090,
                              'autocorr_bern20-03': -72,
                              'chem': -47.70651483,
                              'alkyl': -1.76499965}
        self.primal_sol = dict()
        self.primal_sol['batch'] = _get_sol('batch')
        self.primal_sol['alkyl'] = _get_sol('alkyl')
        self.primal_sol['ball_mk2_10'] = _get_sol('ball_mk2_10')
        self.primal_sol['syn05m'] = _get_sol('syn05m')
        self.primal_sol['autocorr_bern20-03'] = _get_sol('autocorr_bern20-03')
        self.primal_sol['chem'] = _get_sol('chem')
        for pname in self.test_problems.keys():
            get_minlplib(problem_name=pname)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        nlp_solver.config.log_level = logging.DEBUG
        self.opt = coramin.algorithms.MultiTree(mip_solver=mip_solver,
                                                nlp_solver=nlp_solver)

    @classmethod
    def tearDownClass(self) -> None:
        current_dir = os.getcwd()
        for pname in self.test_problems.keys():
            os.remove(os.path.join(current_dir, 'minlplib', 'osil', f'{pname}.osil'))
        os.rmdir(os.path.join(current_dir, 'minlplib', 'osil'))
        os.rmdir(os.path.join(current_dir, 'minlplib'))
        for pname in self.primal_sol.keys():
            os.remove(os.path.join(current_dir, f'{pname}.sol'))

    def get_model(self, pname):
        current_dir = os.getcwd()
        fname = os.path.join(current_dir, 'minlplib', 'osil', f'{pname}.osil')
        m = read_osil(fname, objective_prefix='obj_')
        return m

    def _check_primal_sol(self, pname, m: _BlockData, res: appsi.base.Results):
        expected_by_str = self.primal_sol[pname]
        expected_by_var = pe.ComponentMap()
        for vname, vval in expected_by_str.items():
            v = getattr(m, vname)
            expected_by_var[v] = vval
        got = res.solution_loader.get_primals()
        for v, val in expected_by_var.items():
            self._check_relative_diff(val, got[v])
        got = res.solution_loader.get_primals(vars_to_load=list(expected_by_var.keys()))
        for v, val in expected_by_var.items():
            self._check_relative_diff(val, got[v])

    def optimal_helper(self, pname, check_primal_sol=True):
        m = self.get_model(pname)
        res = self.opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.optimal)
        self._check_relative_diff(self.test_problems[pname], res.best_feasible_objective,
                                  abs_tol=self.opt.config.abs_gap, rel_tol=self.opt.config.mip_gap)
        self._check_relative_diff(self.test_problems[pname], res.best_objective_bound,
                                  abs_tol=self.opt.config.abs_gap, rel_tol=self.opt.config.mip_gap)
        if check_primal_sol:
            self._check_primal_sol(pname, m, res)

    def infeasible_helper(self, pname):
        m = self.get_model(pname)
        self.opt.config.load_solution = False
        res = self.opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.infeasible)
        self.opt.config.load_solution = True

    def time_limit_helper(self, pname):
        orig_time_limit = self.opt.config.time_limit
        self.opt.config.load_solution = False
        for new_limit in [0, 0.1, 0.2, 0.3]:
            self.opt.config.time_limit = new_limit
            m = self.get_model(pname)
            res = self.opt.solve(m)
            self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.maxTimeLimit)
        self.opt.config.load_solution = True

    def test_batch(self):
        self.optimal_helper('batch')

    def test_ball_mk2_10(self):
        self.optimal_helper('ball_mk2_10')

    def test_alkyl(self):
        self.optimal_helper('alkyl')

    def test_syn05m(self):
        self.optimal_helper('syn05m')

    def test_autocorr_bern20_03(self):
        self.optimal_helper('autocorr_bern20-03', check_primal_sol=False)

    def test_chem(self):
        orig_config = self.opt.config()
        self.opt.config.root_obbt_max_iter = 10
        self.opt.config.mip_gap = 0.05
        self.optimal_helper('chem')
        self.opt.config = orig_config

    def test_time_limit(self):
        self.time_limit_helper('alkyl')

    def test_ball_mk3_10(self):
        self.infeasible_helper('ball_mk3_10')

    def test_available(self):
        avail = self.opt.available()
        assert avail in appsi.base.Solver.Availability


class TestMultiTree(Helper):
    def test_convex_overestimator(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(-2, 1))
        m.y = pe.Var()
        m.obj = pe.Objective(expr=(m.x + 1)**2 - 0.2*m.y)
        m.c = pe.Constraint(expr=m.y <= m.x**2)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        nlp_solver.config.log_level = logging.DEBUG
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver,
                                           nlp_solver=nlp_solver)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.optimal)
        self._check_relative_diff(-0.25, res.best_feasible_objective,
                             abs_tol=opt.config.abs_gap, rel_tol=opt.config.mip_gap)
        self._check_relative_diff(-0.25, res.best_objective_bound,
                             abs_tol=opt.config.abs_gap, rel_tol=opt.config.mip_gap)
        self._check_relative_diff(-1.250953, m.x.value, 1e-2, 1e-2)
        self._check_relative_diff(1.5648825, m.y.value, 1e-2, 1e-2)

    def test_max_iter(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(-2, 1))
        m.y = pe.Var()
        m.obj = pe.Objective(expr=(m.x + 1)**2 - 0.2*m.y)
        m.c = pe.Constraint(expr=m.y <= m.x**2)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        nlp_solver.config.log_level = logging.DEBUG
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver,
                                           nlp_solver=nlp_solver)
        opt.config.max_iter = 3
        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.maxIterations)
        self.assertIsNone(res.best_feasible_objective)
        opt.config.max_iter = 10
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.maxIterations)
        self.assertIsNotNone(res.best_feasible_objective)

    def test_nlp_infeas_fbbt(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(-2, 1), domain=pe.Integers)
        m.y = pe.Var(domain=pe.Integers)
        m.obj = pe.Objective(expr=(m.x + 1)**2 - 0.2*m.y)
        m.c1 = pe.Constraint(expr=m.y <= (m.x - 0.5)**2 - 0.5)
        m.c2 = pe.Constraint(expr=m.y >= -(m.x + 2)**2 + 4)
        m.c3 = pe.Constraint(expr=m.y <= 2*m.x + 7)
        m.c4 = pe.Constraint(expr=m.y >= m.x)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        nlp_solver.config.log_level = logging.DEBUG
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver,
                                           nlp_solver=nlp_solver)
        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.infeasible)

    def test_all_vars_fixed_in_nlp(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(-2, 1))
        m.y = pe.Var(domain=pe.Integers)
        m.z = pe.Var()
        m.obj = pe.Objective(expr=m.z - 0.2*m.y)
        m.c1 = pe.Constraint(expr=m.y == (m.x - 0.5)**2 - 0.5)
        m.c2 = pe.Constraint(expr=m.z == (m.x + 1)**2)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        nlp_solver.config.log_level = logging.DEBUG
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver,
                                           nlp_solver=nlp_solver)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.optimal)
        self._check_relative_diff(-0.462486082, res.best_feasible_objective)
        self._check_relative_diff(-0.462486082, res.best_objective_bound)
        self._check_relative_diff(-1.37082869, m.x.value)
        self._check_relative_diff(3, m.y.value)
        self._check_relative_diff(0.137513918, m.z.value)

    def test_linear_problem(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.y = pe.Var()
        m.obj = pe.Objective(expr=m.y)
        m.c1 = pe.Constraint(expr=m.y >= m.x + 1)
        m.c2 = pe.Constraint(expr=m.y >= -m.x + 1)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver, nlp_solver=nlp_solver)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        self.assertAlmostEqual(res.best_objective_bound, 1)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 1)

    def test_stale_fixed_vars(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.y = pe.Var()
        m.z = pe.Var(domain=pe.Binary)
        m.w = pe.Var()
        m.obj = pe.Objective(expr=m.y)
        m.c1 = pe.Constraint(expr=m.y >= m.x + 1)
        m.c2 = pe.Constraint(expr=m.y >= -m.x + 1)
        m.c3 = pe.Constraint(expr=m.w == 2)
        mip_solver = appsi.solvers.Gurobi()
        nlp_solver = appsi.solvers.Ipopt()
        opt = coramin.algorithms.MultiTree(mip_solver=mip_solver, nlp_solver=nlp_solver)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, appsi.base.TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        self.assertAlmostEqual(res.best_objective_bound, 1)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 1)
        self.assertAlmostEqual(m.w.value, 2)
        self.assertIsNone(m.z.value)



from pyomo.core.expr import numeric_expr
from pyomo.core.expr.visitor import identify_variables, ExpressionValueVisitor
from pyomo.core.expr.numvalue import (
    nonpyomo_leaf_types,
    NumericValue,
    is_potentially_variable,
)
from typing import MutableMapping, Tuple, Sequence, Union, List


def _flatten_expr_ProductExpression(
    node: numeric_expr.ProductExpression,
    values: Union[Tuple[NumericValue, ...], List[NumericValue]],
):
    arg1, arg2 = values
    arg1_type = type(arg1)
    arg2_type = type(arg2)
    if is_potentially_variable(arg1) and is_potentially_variable(arg2):
        res = numeric_expr.ProductExpression(values)
    elif is_potentially_variable(arg1):
        if arg1_type is numeric_expr.SumExpression:
            res = numeric_expr.SumExpression([arg2 * i for i in arg1.args])
        elif arg1_type is numeric_expr.LinearExpression:
            res = numeric_expr.LinearExpression(
                constant=arg2 * arg1.constant,
                linear_coefs=[arg2 * i for i in arg1.linear_coefs],
                linear_vars=list(arg1.linear_vars),
            )
        else:
            res = numeric_expr.ProductExpression(values)
    elif is_potentially_variable(arg2):
        if arg2_type is numeric_expr.SumExpression:
            res = numeric_expr.SumExpression([arg1 * i for i in arg2.args])
        elif arg2_type is numeric_expr.LinearExpression:
            res = numeric_expr.LinearExpression(
                constant=arg1 * arg2.constant,
                linear_coefs=[arg1 * i for i in arg2.linear_coefs],
                linear_vars=list(arg2.linear_vars),
            )
        else:
            res = numeric_expr.ProductExpression(values)
    else:
        res = numeric_expr.ProductExpression(values)
    return res


def _flatten_expr_SumExpression(
    node: numeric_expr.SumExpression,
    values: Union[Tuple[NumericValue, ...], List[NumericValue]],
):
    all_args = list()
    for arg in values:
        if isinstance(arg, numeric_expr.SumExpression):
            all_args.extend(arg.args)
        elif isinstance(arg, numeric_expr.LinearExpression):
            for c, v in zip(arg.linear_vars, arg.linear_coefs):
                all_args.append(numeric_expr.MonomialTermExpression((c, v)))
            all_args.append(arg.constant)
        else:
            all_args.append(arg)
    return numeric_expr.SumExpression(all_args)


def _flatten_expr_NegationExpression(
    node: numeric_expr.NegationExpression,
    values: Union[Tuple[NumericValue, ...], List[NumericValue]],
):
    assert len(values) == 1
    arg = values[0]
    if isinstance(arg, numeric_expr.SumExpression):
        res = numeric_expr.SumExpression([-i for i in arg.args])
    elif isinstance(arg, numeric_expr.LinearExpression):
        new_args = [
            numeric_expr.MonomialTermExpression((-c, v))
            for c, v in zip(arg.linear_vars, arg.linear_coefs)
        ]
        new_args.append(-arg.constant)
        res = numeric_expr.SumExpression(new_args)
    else:
        res = numeric_expr.NegationExpression((arg,))
    return res


def _flatten_expr_default(
    node: numeric_expr.ExpressionBase,
    values: Union[Tuple[NumericValue, ...], List[NumericValue]],
):
    return node.create_node_with_local_data(tuple(values))


_flatten_expr_map = dict()
_flatten_expr_map[numeric_expr.SumExpression] = _flatten_expr_SumExpression
_flatten_expr_map[numeric_expr.NegationExpression] = _flatten_expr_NegationExpression
_flatten_expr_map[numeric_expr.ProductExpression] = _flatten_expr_ProductExpression


class FlattenExprVisitor(ExpressionValueVisitor):
    def visit(self, node, values):
        node_type = type(node)
        if node_type in _flatten_expr_map:
            return _flatten_expr_map[node_type](node, values)
        else:
            return _flatten_expr_default(node, values)

    def visiting_potential_leaf(self, node):
        node_type = type(node)
        if node_type in nonpyomo_leaf_types:
            return True, node
        elif not node.is_expression_type():
            return True, node
        elif node_type is numeric_expr.LinearExpression:
            return True, node
        else:
            return False, None


def flatten_expr(expr):
    visitor = FlattenExprVisitor()
    return visitor.dfs_postorder_stack(expr)


class Grouper(object):
    def __init__(self):
        self._terms_by_num_var: MutableMapping[
            int, MutableMapping[Tuple[int, ...], NumericValue]
        ] = dict()

    def add_term(self, expr):
        vlist = list(identify_variables(expr=expr, include_fixed=False))
        vlist.sort(key=lambda x: id(x))
        v_ids = tuple(id(v) for v in vlist)
        num_vars = len(vlist)
        if num_vars not in self._terms_by_num_var:
            self._terms_by_num_var[num_vars] = dict()
        if v_ids not in self._terms_by_num_var[num_vars]:
            self._terms_by_num_var[num_vars][v_ids] = expr
        else:
            self._terms_by_num_var[num_vars][v_ids] += expr

    def group(self) -> Sequence[NumericValue]:
        num_var_list = list(self._terms_by_num_var.keys())
        num_var_list.sort(reverse=True)
        for num_vars in num_var_list[1:]:
            for last_num_vars in num_var_list:
                if last_num_vars == num_vars:
                    break
                for v_ids in list(self._terms_by_num_var[num_vars].keys()):
                    v_id_set = set(v_ids)
                    for last_v_ids in list(
                        self._terms_by_num_var[last_num_vars].keys()
                    ):
                        last_v_id_set = set(last_v_ids)
                        if len(v_id_set - last_v_id_set) == 0:
                            self._terms_by_num_var[last_num_vars][
                                last_v_ids
                            ] += self._terms_by_num_var[num_vars][v_ids]
                            del self._terms_by_num_var[num_vars][v_ids]
                            break

        expr_list = list()
        for num_vars in reversed(num_var_list):
            for e in self._terms_by_num_var[num_vars].values():
                expr_list.append(e)

        return expr_list


def split_expr(expr):
    expr = flatten_expr(expr)
    if type(expr) is numeric_expr.SumExpression:
        grouper = Grouper()
        for arg in expr.args:
            grouper.add_term(arg)
        res = grouper.group()
    else:
        res = [expr]
    return res



import logging
import pyomo.environ as pyo
from coramin.utils.coramin_enums import RelaxationSide
from .custom_block import declare_custom_block
from .relaxations_base import BasePWRelaxationData, ComponentWeakRef, _check_cut
import math
from ._utils import check_var_pts, _get_bnds_list, _get_bnds_tuple
from pyomo.core.base.param import IndexedParam
from pyomo.core.base.constraint import IndexedConstraint
from pyomo.core.expr.numeric_expr import LinearExpression
from typing import Optional, Dict, Sequence
pe = pyo

logger = logging.getLogger(__name__)


def _build_pw_mccormick_relaxation(b, x1, x2, aux_var, x1_pts, relaxation_side=RelaxationSide.BOTH, safety_tol=1e-10):
    """
    This function creates piecewise envelopes to relax "aux_var = x1*x2". Note that the partitioning is done on "x1" only.
    This is the "nf4r" from Gounaris, Misener, and Floudas (2009).

    Parameters
    ----------
    b: pyo.ConcreteModel or pyo.Block
    x1: pyomo.core.base.var._GeneralVarData
        The "x1" variable in x1*x2
    x2: pyomo.core.base.var._GeneralVarData
        The "x2" variable in x1*x2
    aux_var: pyomo.core.base.var._GeneralVarData
        The "aux_var" variable that is replacing x*y
    x1_pts: Sequence[float]
        A list of floating point numbers to define the points over which the piecewise representation will generated.
        This list must be ordered, and it is expected that the first point (x_pts[0]) is equal to x.lb and the
        last point (x_pts[-1]) is equal to x.ub
    relaxation_side : minlp.minlp_defn.RelaxationSide
        Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
    """
    assert len(x1_pts) > 2

    x1_lb = x1_pts[0]
    x1_ub = x1_pts[-1]
    x2_lb, x2_ub = tuple(_get_bnds_list(x2))

    check_var_pts(x1, x_pts=x1_pts)
    check_var_pts(x2)

    if x1.is_fixed() and x2.is_fixed():
        b.x1_x2_fixed_eq = pyo.Constraint(expr= aux_var == pyo.value(x1) * pyo.value(x2))
    elif x1.is_fixed():
        b.x1_fixed_eq = pyo.Constraint(expr= aux_var == pyo.value(x1) * x2)
    elif x2.is_fixed():
        b.x2_fixed_eq = pyo.Constraint(expr= aux_var == x1 * pyo.value(x2))
    else:
        # create the lambda_ variables (binaries for the pw representation)
        b.interval_set = pyo.Set(initialize=range(1, len(x1_pts)))
        b.lambda_ = pyo.Var(b.interval_set, within=pyo.Binary)

        # create the delta x2 variables
        b.delta_x2 = pyo.Var(b.interval_set, bounds=(0, None))

        # create the "sos1" constraint
        b.lambda_sos1 = pyo.Constraint(expr=sum(b.lambda_[n] for n in b.interval_set) == 1.0)

        # create the x1 interval constraints
        b.x1_interval_lb = pyo.Constraint(expr=sum(x1_pts[n - 1] * b.lambda_[n] for n in b.interval_set) <= x1)
        b.x1_interval_ub = pyo.Constraint(expr=x1 <= sum(x1_pts[n] * b.lambda_[n] for n in b.interval_set))

        # create the x2 constraints
        b.x2_con = pyo.Constraint(expr=x2 == x2_lb + sum(b.delta_x2[n] for n in b.interval_set))

        def delta_x2n_ub_rule(m, n):
            return b.delta_x2[n] <= (x2_ub - x2_lb) * b.lambda_[n]

        b.delta_x2n_ub = pyo.Constraint(b.interval_set, rule=delta_x2n_ub_rule)

        # create the relaxation constraints
        if relaxation_side == RelaxationSide.UNDER or relaxation_side == RelaxationSide.BOTH:
            b.aux_var_lb1 = pyo.Constraint(expr=(aux_var >= x2_ub * x1 + sum(x1_pts[n] * b.delta_x2[n] for n in b.interval_set) -
                                                 (x2_ub - x2_lb) * sum(x1_pts[n] * b.lambda_[n] for n in b.interval_set) - safety_tol))
            b.aux_var_lb2 = pyo.Constraint(expr=aux_var >= x2_lb * x1 + sum(x1_pts[n - 1] * b.delta_x2[n] for n in b.interval_set) - safety_tol)

        if relaxation_side == RelaxationSide.OVER or relaxation_side == RelaxationSide.BOTH:
            b.aux_var_ub1 = pyo.Constraint(expr=(aux_var <= x2_ub * x1 + sum(x1_pts[n - 1] * b.delta_x2[n] for n in b.interval_set) -
                                                 (x2_ub - x2_lb) * sum(x1_pts[n - 1] * b.lambda_[n] for n in b.interval_set) + safety_tol))
            b.aux_var_ub2 = pyo.Constraint(expr=aux_var <= x2_lb * x1 + sum(x1_pts[n] * b.delta_x2[n] for n in b.interval_set) + safety_tol)


@declare_custom_block(name='PWMcCormickRelaxation')
class PWMcCormickRelaxationData(BasePWRelaxationData):
    """
    A class for managing McCormick relaxations of bilinear terms (aux_var = x1 * x2).
    """

    def __init__(self, component):
        BasePWRelaxationData.__init__(self, component)
        self._x1ref = ComponentWeakRef(None)
        self._x2ref = ComponentWeakRef(None)
        self._aux_var_ref = ComponentWeakRef(None)
        self._f_x_expr = None
        self._mc_index = None
        self._slopes_index = None
        self._v_index = None
        self._slopes: Optional[IndexedParam] = None
        self._intercepts: Optional[IndexedParam] = None
        self._mccormicks: Optional[IndexedConstraint] = None
        self._mc_exprs: Dict[int, LinearExpression] = dict()
        self._pw = None

    @property
    def _x1(self):
        return self._x1ref.get_component()

    @property
    def _x2(self):
        return self._x2ref.get_component()

    @property
    def _aux_var(self):
        return self._aux_var_ref.get_component()

    def get_rhs_vars(self):
        return self._x1, self._x2

    def get_rhs_expr(self):
        return self._f_x_expr

    def vars_with_bounds_in_relaxation(self):
        return [self._x1, self._x2]

    def _remove_relaxation(self):
        del self._slopes, self._intercepts, self._mccormicks, self._pw, \
            self._mc_index, self._v_index, self._slopes_index
        self._mc_index = None
        self._v_index = None
        self._slopes_index = None
        self._slopes = None
        self._intercepts = None
        self._mccormicks = None
        self._mc_exprs = dict()
        self._pw = None

    def set_input(self, x1, x2, aux_var, relaxation_side=RelaxationSide.BOTH, large_coef=1e5, small_coef=1e-10,
                  safety_tol=1e-10):
        """
        Parameters
        ----------
        x1 : pyomo.core.base.var._GeneralVarData
            The "x1" variable in x1*x2
        x2 : pyomo.core.base.var._GeneralVarData
            The "x2" variable in x1*x2
        aux_var : pyomo.core.base.var._GeneralVarData
            The "aux_var" auxillary variable that is replacing x1*x2
        relaxation_side : minlp.minlp_defn.RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        """
        super(PWMcCormickRelaxationData, self).set_input(relaxation_side=relaxation_side,
                                                         use_linear_relaxation=True,
                                                         large_coef=large_coef, small_coef=small_coef,
                                                         safety_tol=safety_tol)
        self._x1ref.set_component(x1)
        self._x2ref.set_component(x2)
        self._aux_var_ref.set_component(aux_var)
        self._partitions[self._x1] = _get_bnds_list(self._x1)
        self._f_x_expr = x1 * x2

    def build(self, x1, x2, aux_var, relaxation_side=RelaxationSide.BOTH, large_coef=1e5, small_coef=1e-10,
              safety_tol=1e-10):
        """
        Parameters
        ----------
        x1 : pyomo.core.base.var._GeneralVarData
            The "x1" variable in x1*x2
        x2 : pyomo.core.base.var._GeneralVarData
            The "x2" variable in x1*x2
        aux_var : pyomo.core.base.var._GeneralVarData
            The "aux_var" auxillary variable that is replacing x1*x2
        relaxation_side : minlp.minlp_defn.RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        """
        self.set_input(x1=x1, x2=x2, aux_var=aux_var, relaxation_side=relaxation_side,
                       large_coef=large_coef, small_coef=small_coef, safety_tol=safety_tol)
        self.rebuild()

    def remove_relaxation(self):
        super(PWMcCormickRelaxationData, self).remove_relaxation()
        self._remove_relaxation()

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        super(PWMcCormickRelaxationData, self).rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                                                       ensure_oa_at_vertices=ensure_oa_at_vertices)
        if not build_nonlinear_constraint:
            if self._check_valid_domain_for_relaxation():
                if len(self._partitions[self._x1]) == 2:
                    if self._mccormicks is None:
                        self._remove_relaxation()
                        self._build_mccormicks()
                    self._update_mccormicks()
                else:
                    self._remove_relaxation()
                    del self._pw
                    self._pw = pe.Block(concrete=True)
                    _build_pw_mccormick_relaxation(b=self._pw, x1=self._x1, x2=self._x2, aux_var=self._aux_var,
                                                   x1_pts=self._partitions[self._x1],
                                                   relaxation_side=self.relaxation_side, safety_tol=self.safety_tol)
            else:
                self._remove_relaxation()

    def _build_mccormicks(self):
        del self._mc_index, self._v_index, self._slopes_index, self._slopes, self._intercepts, self._mccormicks
        self._mc_exprs = dict()
        self._mc_index = pe.Set(initialize=[0, 1, 2, 3])
        self._v_index = pe.Set(initialize=[1, 2])
        self._slopes_index = pe.Set(initialize=self._mc_index * self._v_index)
        self._slopes = IndexedParam(self._slopes_index, mutable=True)
        self._intercepts = IndexedParam(self._mc_index, mutable=True)
        self._mccormicks = IndexedConstraint(self._mc_index)

        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.UNDER}:
            for ndx in [0, 1]:
                e = LinearExpression(constant=self._intercepts[ndx],
                                     linear_coefs=[self._slopes[ndx, 1], self._slopes[ndx, 2]],
                                     linear_vars=[self._x1, self._x2])
                self._mc_exprs[ndx] = e
                self._mccormicks[ndx] = self._aux_var >= e

        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.OVER}:
            for ndx in [2, 3]:
                e = LinearExpression(constant=self._intercepts[ndx],
                                     linear_coefs=[self._slopes[ndx, 1], self._slopes[ndx, 2]],
                                     linear_vars=[self._x1, self._x2])
                self._mc_exprs[ndx] = e
                self._mccormicks[ndx] = self._aux_var <= e

    def _check_expr(self, ndx):
        if ndx in {0, 1}:
            rel_side = RelaxationSide.UNDER
        else:
            rel_side = RelaxationSide.OVER
        success, bad_var, bad_coef, err_msg = _check_cut(self._mc_exprs[ndx], too_small=self.small_coef,
                                                         too_large=self.large_coef, relaxation_side=rel_side,
                                                         safety_tol=self.safety_tol)
        if not success:
            self._log_bad_cut(bad_var, bad_coef, err_msg)
            self._mccormicks[ndx].deactivate()
        else:
            self._mccormicks[ndx].activate()

    def _update_mccormicks(self):
        x1_lb, x1_ub = _get_bnds_tuple(self._x1)
        x2_lb, x2_ub = _get_bnds_tuple(self._x2)

        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.UNDER}:
            self._slopes[0, 1]._value = x2_lb
            self._slopes[0, 2]._value = x1_lb
            self._intercepts[0]._value = -x1_lb * x2_lb

            self._slopes[1, 1]._value = x2_ub
            self._slopes[1, 2]._value = x1_ub
            self._intercepts[1]._value = -x1_ub * x2_ub

            self._check_expr(0)
            self._check_expr(1)

        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.OVER}:
            self._slopes[2, 1]._value = x2_lb
            self._slopes[2, 2]._value = x1_ub
            self._intercepts[2]._value = -x1_ub * x2_lb

            self._slopes[3, 1]._value = x2_ub
            self._slopes[3, 2]._value = x1_lb
            self._intercepts[3]._value = -x1_lb * x2_ub

            self._check_expr(2)
            self._check_expr(3)

    def add_partition_point(self, value=None):
        """
        This method adds one point to the partitioning of x1. If value is not
        specified, a single point will be added to the partitioning of x1 at the current value of x1. If value is
        specified, then value is added to the partitioning of x1.

        Parameters
        ----------
        value: float
            The point to be added to the partitioning of x1.
        """
        self._add_partition_point(self._x1, value)

    def is_rhs_convex(self):
        """
        Returns True if linear underestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        return False

    def is_rhs_concave(self):
        """
        Returns True if linear overestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        return False



from .obbt import perform_obbt
from .filters import filter_variables_from_solution, aggressive_filter
try:
    from .dbt import decompose_model, perform_dbt, perform_dbt_with_integers_relaxed, TreeBlockData, TreeBlock, \
        DecompositionError, TreeBlockError, collect_vars_to_tighten, collect_vars_to_tighten_by_block, DBTInfo, \
        push_integers, pop_integers, OBBTMethod, FilterMethod
except:
    pass



from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd
from pyomo.core.expr.visitor import identify_variables
from pyomo.common.collections import ComponentSet
import enum
import pyomo.environ as pe
from pyomo.core.expr.numvalue import is_fixed
import math
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
import numpy as np
from coramin.utils.coramin_enums import EigenValueBounder
from pyomo.core.base.block import _BlockData
from typing import Optional, MutableMapping
from pyomo.core.expr.numeric_expr import ExpressionBase
from pyomo.contrib import appsi
from pyomo.common.modeling import unique_component_name
from pyomo.core.base.var import _GeneralVarData
from coramin.utils.pyomo_utils import simplify_expr


def _2d_determinant(mat: np.ndarray):
    return mat[0, 0] * mat[1, 1] - mat[1, 0] * mat[0, 1]


def _determinant(mat):
    nrows, ncols = mat.shape
    assert nrows == ncols
    if nrows == 1:
        det = mat[0, 0]
    elif nrows == 2:
        det = _2d_determinant(mat)
    else:
        i = 0
        det = 0
        next_rows = np.array(list(range(i+1, nrows)), dtype=int)
        for j in range(nrows):
            next_cols = [k for k in range(j)]
            next_cols.extend(k for k in range(j+1, nrows))
            next_cols = np.array(next_cols, dtype=int)
            next_mat = mat[next_rows, :]
            next_mat = next_mat[:, next_cols]
            det += (-1)**(i + j) * mat[i, j] * _determinant(next_mat)
    return simplify_expr(det)


class Hessian(object):
    def __init__(
        self,
        expr: ExpressionBase,
        opt: Optional[appsi.base.Solver],
        method: EigenValueBounder = EigenValueBounder.LinearProgram,
    ):
        self.method = EigenValueBounder(method)
        self.opt = opt
        self._expr = expr
        self._var_list = list(identify_variables(expr=expr, include_fixed=False))
        self._ndx_map = pe.ComponentMap(
            (v, ndx) for ndx, v in enumerate(self._var_list)
        )
        self._hessian = self.compute_symbolic_hessian()
        self._eigenvalue_problem: Optional[_BlockData] = None
        self._eigenvalue_relaxation: Optional[_BlockData] = None
        self._orig_to_relaxation_vars: Optional[
            MutableMapping[_GeneralVarData, _GeneralVarData]
        ] = None

    def variables(self):
        return tuple(self._var_list)

    def formulate_eigenvalue_problem(self, sense=pe.minimize):
        if self._eigenvalue_problem is not None:
            min_eig, max_eig = self.bound_eigenvalues_from_interval_hessian()
            if min_eig > self._eigenvalue_problem.eig.lb:
                self._eigenvalue_problem.eig.setlb(min_eig)
            if max_eig < self._eigenvalue_problem.eig.ub:
                self._eigenvalue_problem.eig.setub(max_eig)
            self._eigenvalue_problem.obj.sense = sense
            return self._eigenvalue_problem
        min_eig, max_eig = self.bound_eigenvalues_from_interval_hessian()
        m = pe.ConcreteModel()
        m.eig = pe.Var(bounds=(min_eig, max_eig))
        m.obj = pe.Objective(expr=m.eig, sense=sense)
        for v in self._var_list:
            m.add_component(v.name, pe.Reference(v))

        n = len(self._var_list)
        np_hess = np.empty((n, n), dtype=object)
        for ndx1, v1 in enumerate(self._var_list):
            hess_v1 = self._hessian[v1]
            for ndx2 in range(ndx1, n):
                v2 = self._var_list[ndx2]
                if v2 in hess_v1:
                    np_hess[ndx1, ndx2] = hess_v1[v2]
                else:
                    np_hess[ndx1, ndx2] = 0
                if v1 is v2:
                    np_hess[ndx1, ndx2] -= m.eig
                else:
                    np_hess[ndx2, ndx1] = np_hess[ndx1, ndx2]
        m.det_con = pe.Constraint(expr=_determinant(np_hess) == 0)
        self._eigenvalue_problem = m
        return m

    def formulate_eigenvalue_relaxation(self, sense=pe.minimize):
        if self._eigenvalue_relaxation is not None:
            for orig_v, rel_v in self._orig_to_relaxation_vars.items():
                orig_lb, orig_ub = orig_v.bounds
                rel_lb, rel_ub = rel_v.bounds
                if orig_lb is not None:
                    if rel_lb is None or orig_lb > rel_lb:
                        rel_v.setlb(orig_lb)
                if orig_ub is not None:
                    if rel_ub is None or orig_ub < rel_ub:
                        rel_v.setub(orig_ub)
            from .iterators import relaxation_data_objects
            for b in relaxation_data_objects(self._eigenvalue_relaxation, descend_into=True, active=True):
                b.rebuild()
            self._eigenvalue_relaxation.obj.sense = sense
            return self._eigenvalue_relaxation
        m = self.formulate_eigenvalue_problem(sense=sense)
        all_vars = list(
            ComponentSet(
                m.component_data_objects(pe.Var, descend_into=True)
            )
        )
        tmp_name = unique_component_name(m, "all_vars")
        setattr(m, tmp_name, all_vars)
        from .auto_relax import relax
        relaxation = relax(m, in_place=False)
        new_vars = getattr(relaxation, "all_vars")
        self._orig_to_relaxation_vars = pe.ComponentMap(zip(all_vars, new_vars))
        delattr(m, tmp_name)
        delattr(relaxation, tmp_name)
        self._eigenvalue_relaxation = relaxation
        return relaxation

    def get_minimum_eigenvalue(self):
        if self.method <= EigenValueBounder.GershgorinWithSimplification:
            res = self.bound_eigenvalues_from_interval_hessian()[0]
        elif self.method == EigenValueBounder.LinearProgram:
            m = self.formulate_eigenvalue_relaxation()
            res = self.opt.solve(m).best_objective_bound
        else:
            m = self.formulate_eigenvalue_problem()
            res = self.opt.solve(m).best_objective_bound
        return res

    def get_maximum_eigenvalue(self):
        if self.method <= EigenValueBounder.GershgorinWithSimplification:
            res = self.bound_eigenvalues_from_interval_hessian()[1]
        elif self.method == EigenValueBounder.LinearProgram:
            m = self.formulate_eigenvalue_relaxation(sense=pe.maximize)
            res = self.opt.solve(m).best_objective_bound
        else:
            m = self.formulate_eigenvalue_problem(sense=pe.maximize)
            res = self.opt.solve(m).best_objective_bound
        return res

    def bound_eigenvalues_from_interval_hessian(self):
        ih = self.compute_interval_hessian()
        min_eig = math.inf
        max_eig = -math.inf
        for v1 in self._var_list:
            h = ih[v1]
            if v1 in h:
                row_min = h[v1][0]
                row_max = h[v1][1]
            else:
                row_min = 0
                row_max = 0
            for v2, (lb, ub) in h.items():
                if v2 is v1:
                    continue
                row_min -= max(abs(lb), abs(ub))
                row_max += max(abs(lb), abs(ub))
            min_eig = min(min_eig, row_min)
            max_eig = max(max_eig, row_max)
        return min_eig, max_eig

    def compute_symbolic_hessian(self):
        ders = reverse_sd(self._expr)
        ders2 = pe.ComponentMap()
        for v in self._var_list:
            ders2[v] = reverse_sd(ders[v])

        res = pe.ComponentMap()
        for v in self._var_list:
            res[v] = pe.ComponentMap()

        n = len(self._var_list)
        for v1 in self._var_list:
            ndx1 = self._ndx_map[v1]
            for ndx2 in range(ndx1, n):
                v2 = self._var_list[ndx2]
                if v2 not in ders2[v1]:
                    continue
                der = ders2[v1][v2]
                if is_fixed(der):
                    val = pe.value(der)
                    res[v1][v2] = val
                else:
                    if self.method >= EigenValueBounder.GershgorinWithSimplification:
                        _der = simplify_expr(der)
                    else:
                        _der = der
                    res[v1][v2] = _der
                res[v2][v1] = res[v1][v2]

        return res

    def compute_interval_hessian(self):
        res = pe.ComponentMap()
        for v in self._var_list:
            res[v] = pe.ComponentMap()

        n = len(self._var_list)
        for ndx1, v1 in enumerate(self._var_list):
            for ndx2 in range(ndx1, n):
                v2 = self._var_list[ndx2]
                if v2 not in self._hessian[v1]:
                    continue
                lb, ub = compute_bounds_on_expr(self._hessian[v1][v2])
                if lb is None:
                    lb = -math.inf
                if ub is None:
                    ub = math.inf
                res[v1][v2] = (lb, ub)
                res[v2][v1] = (lb, ub)
        return res

    def pprint(self, intervals=False):
        if intervals:
            ih = self.compute_interval_hessian()
        else:
            ih = self._hessian
        n = len(self._var_list)
        lens = np.ones((n, n), dtype=int)
        strs = dict()
        for ndx in range(n):
            strs[ndx] = dict()

        for v1 in self._var_list:
            ndx1 = self._ndx_map[v1]
            for ndx2 in range(ndx1, n):
                v2 = self._var_list[ndx2]
                if v2 in ih[v1]:
                    der = ih[v1][v2]
                else:
                    if intervals:
                        der = (0, 0)
                    else:
                        der = 0
                if intervals:
                    lb, ub = der
                    der_str = f"({lb:<.3f}, {ub:<.3f})"
                else:
                    der_str = str(der)
                strs[ndx1][ndx2] = der_str
                strs[ndx2][ndx1] = der_str
                lens[ndx1, ndx2] = len(der_str)
                lens[ndx2, ndx1] = len(der_str)

        col_lens = np.max(lens, axis=0)
        row_string = ""
        for ndx, cl in enumerate(col_lens):
            row_string += f"{{{ndx}:<{cl+2}}}"

        res = ""
        for row_ndx in range(n):
            row_entries = tuple(strs[row_ndx][i] for i in range(n))
            res += row_string.format(*row_entries)
            res += '\n'

        print(res)



import os
from zipfile import ZipFile
from pyomo.common import fileutils
from pyomo.common import download
import enum
import math
import csv
from collections.abc import Iterable
import logging


logger = logging.getLogger(__name__)


def get_minlplib_instancedata(target_filename=None):
    """
    Download instancedata.csv from MINLPLib which can be used to get statistics on the problems from minlplib.

    Parameters
    ----------
    target_filename: str
        The full path, including the filename for where to place the downloaded 
        file. The default will be a directory called minlplib in the current 
        working directory and a filename of instancedata.csv.
    """
    if target_filename is None:
        target_filename = os.path.join(os.getcwd(), 'minlplib', 'instancedata.csv')
    download_dir = os.path.dirname(target_filename)

    if os.path.exists(target_filename):
        raise ValueError('A file named {filename} already exists.'.format(filename=target_filename))
    
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    downloader = download.FileDownloader()
    downloader.set_destination_filename(target_filename)
    downloader.get_text_file('http://www.minlplib.org/instancedata.csv')


def _process_acceptable_arg(name, arg, default):
    if arg is None:
        return default
    if isinstance(arg, str):
        if arg not in default:
            raise ValueError("Unrecognized argument for %s: %s" % (name, arg))
        return set([arg])
    if isinstance(arg, Iterable):
        ans = set(str(_) for _ in arg)
        if not ans.issubset(default):
            unknown = default - ans
            raise ValueError("Unrecognized argument for %s: %s" % (name, unknown))
        return ans
    if type(arg) == bool:
        if str(arg) in default:
            return set([str(arg)])
    raise ValueError('unrecognized type for %s: %s' % (name, type(arg)))


def _check_int_arg(arg, _min, _max, arg_name, case_name):
    if arg < _min or arg > _max:
        logger.debug('excluding {case_name} due to {arg_name}'.format(case_name=case_name, arg_name=arg_name))
        return True
    return False


def _check_acceptable(arg, acceptable_set, arg_name, case_name):
    if arg not in acceptable_set:
        logger.debug('excluding {case_name} due to {arg_name}'.format(case_name=case_name, arg_name=arg_name))
        return True
    return False


def filter_minlplib_instances(instancedata_filename=None,
                              min_nvars=0, max_nvars=math.inf,
                              min_nbinvars=0, max_nbinvars=math.inf,
                              min_nintvars=0, max_nintvars=math.inf,
                              min_nnlvars=0, max_nnlvars=math.inf,
                              min_nnlbinvars=0, max_nnlbinvars=math.inf,
                              min_nnlintvars=0, max_nnlintvars=math.inf,
                              min_nobjnz=0, max_nobjnz=math.inf,
                              min_nobjnlnz=0, max_nobjnlnz=math.inf,
                              min_ncons=0, max_ncons=math.inf,
                              min_nlincons=0, max_nlincons=math.inf,
                              min_nquadcons=0, max_nquadcons=math.inf,
                              min_npolynomcons=0, max_npolynomcons=math.inf,
                              min_nsignomcons=0, max_nsignomcons=math.inf,
                              min_ngennlcons=0, max_ngennlcons=math.inf,
                              min_njacobiannz=0, max_njacobiannz=math.inf,
                              min_njacobiannlnz=0, max_njacobiannlnz=math.inf,
                              min_nlaghessiannz=0, max_nlaghessiannz=math.inf,
                              min_nlaghessiandiagnz=0, max_nlaghessiandiagnz=math.inf,
                              min_nsemi=0, max_nsemi=math.inf,
                              min_nnlsemi=0, max_nnlsemi=math.inf,
                              min_nsos1=0, max_nsos1=math.inf,
                              min_nsos2=0, max_nsos2=math.inf,
                              acceptable_formats=None,
                              acceptable_probtype=None,
                              acceptable_objtype=None,
                              acceptable_objcurvature=None,
                              acceptable_conscurvature=None,
                              acceptable_convex=None):
    """
    This function filters problems from MINLPLib based on
    instancedata.csv from MINLPLib and the conditions specified
    through the function arguments. The function argument names
    correspond to column headings from instancedata.csv. The 
    arguments starting with min or max require int or float inputs. 
    The arguments starting with acceptable require either a 
    string or an iterable of strings. See the MINLPLib documentation 
    for acceptable values.
    """
    if instancedata_filename is None:
        instancedata_filename = os.path.join(os.getcwd(), 'minlplib', 'instancedata.csv')

    if not os.path.exists(instancedata_filename):
        raise RuntimeError('{filename} does not exist. Please use get_minlplib_instancedata() first or specify the location of the MINLPLib instancedata.csv with the instancedata_filename argument.'.format(filename=instancedata_filename))

    acceptable_formats = _process_acceptable_arg(
        'acceptable_formats',
        acceptable_formats,
        set(['ams', 'gms', 'lp', 'mod', 'nl', 'osil', 'pip'])
    )

    default_acceptable_probtype = set()
    for pre in ['B', 'I', 'MI', 'MB', 'S', '']:
        for post in ['NLP', 'QCQP', 'QP', 'QCP', 'P']:
            default_acceptable_probtype.add(pre + post)
    acceptable_probtype = _process_acceptable_arg(
        'acceptable_probtype',
        acceptable_probtype,
        default_acceptable_probtype
        )

    acceptable_objtype = _process_acceptable_arg(
        'acceptable_objtype',
        acceptable_objtype,
        set(['constant', 'linear', 'quadratic', 'polynomial', 'signomial', 'nonlinear'])
        )

    acceptable_objcurvature = _process_acceptable_arg(
        'acceptable_objcurvature',
        acceptable_objcurvature,
        set(['linear', 'convex', 'concave', 'indefinite', 'nonconvex', 'nonconcave', 'unknown'])
        )

    acceptable_conscurvature = _process_acceptable_arg(
        'acceptable_conscurvature',
        acceptable_conscurvature,
        set(['linear', 'convex', 'concave', 'indefinite', 'nonconvex', 'nonconcave', 'unknown'])
        )

    acceptable_convex = _process_acceptable_arg(
        'acceptable_convex',
        acceptable_convex,
        set(['True', 'False', ''])
        )

    int_arg_name_list = [
        'nvars',
        'nbinvars',
        'nintvars',
        'nnlvars',
        'nnlbinvars',
        'nnlintvars',
        'nobjnz',
        'nobjnlnz',
        'ncons',
        'nlincons',
        'nquadcons',
        'npolynomcons',
        'nsignomcons',
        'ngennlcons',
        'njacobiannz',
        'njacobiannlnz',
        'nlaghessiannz',
        'nlaghessiandiagnz',
        'nsemi',
        'nnlsemi',
        'nsos1',
        'nsos2',
    ]
    min_list = [
        min_nvars,
        min_nbinvars,
        min_nintvars,
        min_nnlvars,
        min_nnlbinvars,
        min_nnlintvars,
        min_nobjnz,
        min_nobjnlnz,
        min_ncons,
        min_nlincons,
        min_nquadcons,
        min_npolynomcons,
        min_nsignomcons,
        min_ngennlcons,
        min_njacobiannz,
        min_njacobiannlnz,
        min_nlaghessiannz,
        min_nlaghessiandiagnz,
        min_nsemi,
        min_nnlsemi,
        min_nsos1,
        min_nsos2,
    ]
    max_list = [
        max_nvars,
        max_nbinvars,
        max_nintvars,
        max_nnlvars,
        max_nnlbinvars,
        max_nnlintvars,
        max_nobjnz,
        max_nobjnlnz,
        max_ncons,
        max_nlincons,
        max_nquadcons,
        max_npolynomcons,
        max_nsignomcons,
        max_ngennlcons,
        max_njacobiannz,
        max_njacobiannlnz,
        max_nlaghessiannz,
        max_nlaghessiandiagnz,
        max_nsemi,
        max_nnlsemi,
        max_nsos1,
        max_nsos2,
    ]

    acceptable_arg_name_list = [
        'probtype',
        'objtype',
        'objcurvature',
        'conscurvature',
        'convex'
    ]
    acceptable_set_list = [
        acceptable_probtype,
        acceptable_objtype,
        acceptable_objcurvature,
        acceptable_conscurvature,
        acceptable_convex
    ]

    with open(instancedata_filename, 'r') as csv_file:
        reader = csv.reader(csv_file, delimiter=';')
        headings = {column: ndx for ndx, column in enumerate(next(reader))}
        rows = [row for row in reader]

    cases = list()
    for ndx, row in enumerate(rows):
        if len(row) == 0:
            continue
        
        case_name = row[headings['name']]

        available_formats = row[headings['formats']]
        available_formats = available_formats.replace('set([', '')
        available_formats = available_formats.replace('])', '')
        available_formats = available_formats.replace('{', '')
        available_formats = available_formats.replace('}', '')
        available_formats = available_formats.replace(' ', '')
        available_formats = available_formats.replace("'", '')
        available_formats = available_formats.split(',')
        available_formats = set(available_formats)

        should_continue = False

        if len(acceptable_formats.intersection(available_formats)) == 0:
            logger.debug('excluding {case} due to available_formats'.format(case=case_name))
            should_continue = True

        for ndx, acceptable_arg_name in enumerate(acceptable_arg_name_list):
            acceptable_set = acceptable_set_list[ndx]
            arg = row[headings[acceptable_arg_name]]
            if _check_acceptable(arg=arg,
                                 acceptable_set=acceptable_set,
                                 arg_name=acceptable_arg_name,
                                 case_name=case_name):
                should_continue = True

        for ndx, arg_name in enumerate(int_arg_name_list):
            _min = min_list[ndx]
            _max = max_list[ndx]
            arg = int(row[headings[arg_name]])
            if _check_int_arg(arg=arg,
                              _min=_min,
                              _max=_max,
                              arg_name=arg_name,
                              case_name=case_name):
                should_continue = True

        if should_continue:
            continue

        cases.append(case_name)

    return cases


def get_minlplib(download_dir=None, format='osil', problem_name=None):
    """
    Download MINLPLib

    Parameters
    ----------
    download_dir: str
        The directory in which to place the downloaded files. The default will be a 
        current_working_directory/minlplib/file_format/.
    format: str
        The file format requested. Options are ams, gms, lp, mod, nl, osil, and pip
    problem_name: None or str
        If problem_name is None, then the entire zip file will be downloaded 
        and extracted (all problems with the specified format). If a single problem 
        needs to be downloaded, then the name of the problem can be specified. 
        This can be significantly faster than downloading all of the problems. 
        However, individual problems are not compressed, so downloading multiple 
        individual problems can quickly become expensive.
    """
    if download_dir is None:
        download_dir = os.path.join(os.getcwd(), 'minlplib', format)

    if problem_name is None:
        if os.path.exists(download_dir):
            raise ValueError(
                'The specified download_dir already exists: ' + download_dir)

        os.makedirs(download_dir)
        downloader = download.FileDownloader()
        zip_dirname = os.path.join(download_dir, 'minlplib_'+format)
        downloader.set_destination_filename(zip_dirname)
        downloader.get_zip_archive('http://www.minlplib.org/minlplib_'+format+'.zip')
        for i in os.listdir(os.path.join(download_dir, 'minlplib_'+format, 'minlplib', format)):
            os.rename(os.path.join(download_dir, 'minlplib_'+format, 'minlplib', format, i), os.path.join(download_dir, i))
        os.rmdir(os.path.join(download_dir, 'minlplib_'+format, 'minlplib', format))
        os.rmdir(os.path.join(download_dir, 'minlplib_'+format, 'minlplib'))
        os.rmdir(os.path.join(download_dir, 'minlplib_'+format))
    else:
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)
        target_filename = os.path.join(download_dir, problem_name + '.' + format)
        if os.path.exists(target_filename):
            raise ValueError(f'The target filename ({target_filename}) already exists')
        downloader = download.FileDownloader()
        downloader.set_destination_filename(target_filename)
        downloader.get_binary_file('http://www.minlplib.org/'+format+'/'+problem_name+'.'+format)
        
